"""
@author: Jan Tuziemski, Filip Maciejewski, Joanna Majsak, Oskar Słowik, Marcin Kotowski, Katarzyna Kowalczyk-Murynka, Paweł Przewłocki, Piotr Podziemski, Michał Oszmaniec
@contact: filip.b.maciejewski@gmail.com
"""

import copy
import itertools

import numpy as np
import picos
from scipy import optimize as scopt


from qrem.functions_qrem import ancillary_functions as anf
from qrem.functions_qrem import povmtools
from qrem.noise_mitigation.expected_values.CRS.scipy_optimizers_wrappers import \
    wrapper_scipy_optimize as scop_wrapper
# from qiskit.quantum_info import partial_trace as qiskit_partial_trace
# from qiskit.quantum_info import DensityMatrix as qiskit_DM

from qrem.common.math import Constants as Const
# from qrem.common.printer import qprint, qprint_array

__worst_case_distance_names__ = ['operational', 'worst', 'worst_case', 'worst-case', 'wc']
__average_case_distance_names__ = ['average', 'average_case', 'average-case', 'ac']


def operational_distance_POVMs(POVM1,
                               POVM2=None,
                               method='direct',
                               classical=False):

    #JT: computation of classical operational distance is performed

    if classical:
        noise_map_M, noise_map_P = povmtools.get_stochastic_map_from_povm(
            POVM1), povmtools.get_stochastic_map_from_povm(POVM2)
        return 1 / 2 * np.linalg.norm(noise_map_M - noise_map_P, ord=1)

    #JT: calculation of quantum worst case distance starts

    #JT: number of effects of POVM! is set to m

    m = len(POVM1)

    #JT: If POVM2 is not specified it is set to idel computational basis projective measurement

    if POVM2 is None:

        difference = []
        for i in range(m):
            diff_now = copy.deepcopy(POVM1[i])
            diff_now[i, i] -= 1
            difference.append(diff_now)
        difference = (difference)

    #JT: If POVM2 is speciffied differnce in effects is computed

    else:

        difference = ([POVM1[i] - POVM2[i] for i in range(m)])


    #JT: for CBTN computation is performed as below


    if (method == 'CBTN'):

        n = POVM1[0].shape[0]

        #JT: Choi-Jamiolkowski state corresponding to difference of POVMs is computed
        J = povmtools.get_choi_from_POVM(difference)

        #JT: old comment below:
        # this calculates completely bounded trace norm of the channel which is the upper bound for operational distance
        cbt_norm = povmtools.get_CBT_norm(J, n, m)

        return cbt_norm / 2


    #JT: defult computation method

    elif (method == 'direct'):
        # calculate operational distance directly via bruteforce search over subsets_list of outcomes
        biggest_norm = 0

        #JT: k goes from number of effects m to 0

        for k in list(range(m))[::-1]:

            #JT: combinations of k elements from all effects diferrences are created
            #k starts from number of effects

            current_list = list(itertools.combinations(difference, k + 1))

            #JT: tis loop goes over all elements forming a particulat combination

            for l in current_list:
                current_sum = sum(l)

                #JT: Norm of sum of differences is computed

                current_norm = np.linalg.norm(current_sum, ord=2)

                #JT: a check whether the currectly computed norm is higher than the previous ones

                if (current_norm > biggest_norm):
                    biggest_norm = current_norm

        return biggest_norm


#JT: This function is used to compute average case distance between POVMs

def average_distance_POVMs(POVM1,
                           POVM2=None,
                           classical=False):
    """
    Description:
         Get average (over all quantum states) operational distance between two povms
         COMMENT: Assuming that both POVMs have the same number of outcomes
    Parameter:
        :param POVM1: list of arrays representing effects of POVM
        :param POVM2: list of arrays representing effects of POVM, if None assuming it's computational basis
    Return:
        average (over all quantum states) operational distance between two povms
    """

    #JT: classical is a boolean flag determining if av case calassical distance is computed

    if classical:

        #JT:Stochastic noise acting on POVM1 and POVM2 are established

        noise_map_M, noise_map_P = povmtools.get_stochastic_map_from_povm(
            POVM1), povmtools.get_stochastic_map_from_povm(POVM2)

        #JT: average case distance is computed and its value is returned

        return 1 / 2 * np.sqrt(
            1 / 2 * np.linalg.norm(noise_map_M-noise_map_P, ord='fro') ** 2 + np.trace(
                noise_map_M - noise_map_P) ** 2)

    #JT: calculatioof quantum distance starts

    #JT: number of rows of POVM1 is set as dimension

    dimension = POVM1[0].shape[0]

    #JT: number of effects is set as n

    n = len(POVM1)

    dist = 0

    #JT a loop over effects

    for i in range(n):

        #A difference between POVMs effects is computed

        if POVM2 is not None:
            Di = POVM1[i] - POVM2[i]

        #JT: If POVM2 is not specified at input it is assumed that it corresponds to computational basis measurement
        #the difference between effects is then computed by subtracting 1 from diagonal element of POVM1

        else:
            # assuming comp basis
            Di = copy.deepcopy(POVM1[i])
            Di[i, i] -= 1
        # print(Di)

        #JT:Value of distace is updated

        dist += np.sqrt((np.trace(Di) ** 2).real + (np.trace(Di @ Di)).real)


    #

    # Factor 1/2 is okay
    return 1 / dimension * dist / 2


#JT: This function is used to calculate distances between POVMs
#Usage 1: Computation between reconstructed POVMs and thier ideal projective measurements counterparts

def calculate_distance_between_POVMs(POVM_1,
                                     POVM_2,
                                     distance_type_tuple=('average_case', 'classical')):
    distance_type = distance_type_tuple[0]
    errors_type = distance_type_tuple[1]

    classical_errors = False
    if errors_type.lower() in ['classical']:
        classical_errors = True

    #JT: __average_case_distance_names__ - a constant holding  references to average case distance
    #Here computation of av case distance is performed

    if distance_type in __average_case_distance_names__:
        distance_calculated = average_distance_POVMs(POVM1=POVM_1,
                                                     POVM2=POVM_2,
                                                     classical=classical_errors)


    #JT:  __worst_case_distance_names__ - a constant holding  references to worst case distance
    #Here computation of wc case distance is performed

    elif distance_type in __worst_case_distance_names__:
        distance_calculated = operational_distance_POVMs(POVM1=POVM_1,
                                                         POVM2=POVM_2,
                                                         classical=classical_errors)

    #JT: error message when wrong parameters are passed

    else:
        raise ValueError(f"Wrong distance type: {distance_type_tuple}")

    return distance_calculated



def __swap_2q_povm(povm_2q_input):
    povm_2q = [Const.standard_gates()['SWAP'] @ Mi @ Const.standard_gates()['SWAP'] for Mi in
               povm_2q_input]
    povm_2q[1], povm_2q[2] = povm_2q[2], povm_2q[1]

    return povm_2q


def __swap_or_not(povm_2q_input,
                  target_qubit=0):
    if target_qubit == 0:
        povm_2q = copy.deepcopy(povm_2q_input)

    elif target_qubit == 1:
        povm_2q = __swap_2q_povm(povm_2q_input=povm_2q_input)

    else:
        raise ValueError(f"Index of qubit '{target_qubit}' is incorrect, should be 0 or 1.")

    return povm_2q


def __solve_subproblem_average_case(povm_2q_input,
                                    signs_bitstring,
                                    target_qubit
                                    ):
    signs = [(-1) ** int(s) for s in list(signs_bitstring)]

    povm_2q = __swap_or_not(povm_2q_input=povm_2q_input,
                            target_qubit=target_qubit)

    problem = picos.Problem()

    local_difference = picos.HermitianVariable(name='local_difference',
                                               shape=(2, 2))
    problem.add_constraint(picos.trace(local_difference) == 0)
    problem.add_constraint(local_difference << np.eye(2))
    problem.add_constraint(local_difference >> -np.eye(2))

    Madded = povm_2q[0] + povm_2q[1]

    # Madded2 = povm_2q[2]+povm_2q[3]

    # raise KeyboardInterrupt

    M00 = Madded[0:2, 0:2]
    M01 = Madded[0:2, 2:4]
    M10 = Madded[2:4, 0:2]
    M11 = Madded[2:4, 2:4]

    # Ms = [M00, M01, M11]

    # first_part = picos.sum([picos.trace(local_difference*Mi)**2 for Mi in Ms])

    t00 = signs[0] * picos.trace(local_difference * M00)
    t01 = signs[1] * picos.trace(local_difference * (M01 + M10))
    t11 = signs[2] * picos.trace(local_difference * M11)

    t_add = signs[3] * picos.trace((M00 + M11) * local_difference)
    cost_function = t00 + t01 + t11 + t_add

    # print(cost_function)
    problem.set_objective(direction='max',
                          expression=cost_function)

    # print(problem)
    problem.solve(solver='mosek')

    # print(cost_function.value)

    # print('indirect:')
    # best_local_difference = np.array(local_difference.value_as_matrix)
    # qprint_array(best_local_difference)
    # print(np.round(np.linalg.eigvals(best_local_difference),3))

    dav = 1 / 2 * np.sqrt(
        t00.coherences_values ** 2 + t01.coherences_values ** 2 + t11.coherences_values ** 2 + t_add.coherences_values ** 2)

    # qprint_array(np.array(local_difference.value_as_matrix))
    return dav.real


def calculate_correlations_coefficients_average_case(povm_2q):
    full_register = anf.get_classical_register_bitstrings(qubit_indices=range(4),
                                                          quantum_register_size=4,
                                                          )

    solutions_list_i_j, solutions_list_j_i = [], []
    for signs_bitstring in full_register:
        dav_now_i_j = __solve_subproblem_average_case(povm_2q_input=povm_2q,
                                                      target_qubit=0,
                                                      signs_bitstring=signs_bitstring)
        solutions_list_i_j.append(dav_now_i_j)

        dav_now_j_i = __solve_subproblem_average_case(povm_2q_input=povm_2q,
                                                      target_qubit=1,
                                                      signs_bitstring=signs_bitstring)
        solutions_list_j_i.append(dav_now_j_i)

    # how "i" is affected by "j"
    c_i_j = np.max(solutions_list_i_j)

    # how "j is affected by "i"
    c_j_i = np.max(solutions_list_j_i)

    # print(solutions_list_i_j)
    # print(solutions_list_j_i)

    return c_i_j, c_j_i


def _solve_max_dop_problem_for_fixed_input_state(input_state,
                                                 povm_2q,
                                                 sign):
    # print('hejunia')
    problem = picos.Problem()

    local_difference = picos.HermitianVariable(name='local_difference',
                                               shape=(2, 2))
    problem.add_constraint(picos.trace(local_difference) == 0)
    problem.add_constraint(local_difference << np.eye(2))
    problem.add_constraint(local_difference >> -np.eye(2))

    embedded_difference = picos.kron(np.eye(2), local_difference)

    local_effect = picos.partial_trace((povm_2q[0] + povm_2q[1]) * embedded_difference,
                                       subsystems=1,
                                       dimensions=2)

    cost_function = sign * picos.trace(input_state * local_effect)

    problem.set_objective(direction='max',
                          expression=cost_function)

    # print(problem)

    problem.solve(solver='mosek')

    dop = cost_function.coherences_values

    return dop.real


# from povms_qi import povmtools as pt
# povm_2q = pt.random_POVM(4,2)
# sign=1


def __construct_local_difference(parameters):
    nx, ny, nz = parameters

    local_difference = nx * Const.pauli_sigmas()['X'] + ny * Const.pauli_sigmas()['Y'] + nz * Const.pauli_sigmas()[
        'Z']
    return local_difference


def _cost_function_difference_dac(parameters_difference,
                                  Ms_list):
    nx, ny, nz = parameters_difference[0], parameters_difference[1], parameters_difference[2]
    local_difference = nx * Const.pauli_sigmas()['X'] + ny * Const.pauli_sigmas()['Y'] + nz * Const.pauli_sigmas()[
        'Z']

    M00, M01, M10, M11 = Ms_list
    a = np.trace(local_difference @ M00).real
    b = np.trace(local_difference @ M11).real
    c = np.trace(M01 @ local_difference)
    d = np.trace(M10 @ local_difference)

    # c = np.trace((M01+M10) @ local_difference)
    # d = 0

    trace_part = a + b

    norm_squared = a ** 2 + b ** 2 + c ** 2 + d ** 2 + trace_part ** 2
    norm_squared = norm_squared.real
    # # print(norm_squared)
    # a00 = np.trace(local_difference @ M00).real
    # a01 = np.trace(local_difference @ M01)
    # a10 = np.conj(a01)
    # a10 = np.trace(local_difference @ M10)
    #     # np.trace(local_difference @ (M01 + M10)).real
    # a11 = np.trace(local_difference@M11).real
    # #
    # #
    # reduced_effect = np.array([[a00,a01],
    #                              [a10,a11]])
    #
    #
    # norm_squared_true = np.trace(reduced_effect@reduced_effect).real+(np.trace(reduced_effect).real)**2
    #
    # print(norm_squared,norm_squared_true)
    # # print(np.allclose(norm_squared, norm_squared_true))
    # #
    # # print(norm_squared,np.linalg.norm(reduced_effect,ord='fro')**2)
    # #
    # # Madded = np.zeros((4,4),dtype=complex)
    # #
    # # Madded[0:2, 0:2] = M00[:,:]
    # # Madded[0:2, 2:4] = M01[:,:]
    # # Madded[2:4, 0:2] = M10[:,:]
    # # Madded[2:4, 2:4] = M11[:,:]
    # #
    # #
    # # # qprint_array(M)
    # # test = qiskit_DM(data=Madded, dims=4)
    # #
    # # product_test = Madded@np.kron(np.eye(2),local_difference)
    # #
    # #
    # # reduced_effect_test = qiskit_partial_trace(product_test,qargs=[0]).data
    # # #
    # # #
    # #
    # # print(np.allclose(reduced_effect,reduced_effect_test))
    #
    # # qprint_array(reduced_effect)
    # # qprint_array(reduced_effect_test)
    #
    # raise KeyError

    return -norm_squared


def _cost_function_difference_dwc(parameters_difference,
                                  Ms_list):
    nx, ny, nz = parameters_difference[0], parameters_difference[1], parameters_difference[2]

    # norm_check = nx**2+ny ** 2 + nz ** 2
    #
    # if norm_check>1.0:
    #     raise ValueError("damn")
    #     return 100*norm_check

    local_difference = nx * Const.pauli_sigmas()['X'] + ny * Const.pauli_sigmas()['Y'] + nz * Const.pauli_sigmas()[
        'Z']

    M00, M01, M10, M11 = Ms_list
    a00 = np.trace(local_difference @ M00).real
    a01 = np.trace(local_difference @ M01)
    # a10 = np.conj(a01)
    a10 = np.trace(local_difference @ M10)
    # np.trace(local_difference @ (M01 + M10)).real
    a11 = np.trace(local_difference @ M11).real
    #
    # det = a00*a10-a01*a10
    # mean = (a00+a11)/2
    #
    # second_part = np.sqrt(mean**2-det)

    operator_norm = np.linalg.norm(np.array([[a00, a01],
                                             [a10, a11]]),
                                   ord=2).real

    return -operator_norm


#
# def _direct_optimization_dac(povm2q,
#                              target_qubit,
#                              printing=False):
def direct_optimization_difference(povm2q,
                                   target_qubit,
                                   distance_type,
                                   printing=False):
    povm_2q = __swap_or_not(povm_2q_input=povm2q,
                            target_qubit=target_qubit)

    Madded = povm_2q[0] + povm_2q[1]
    M00 = Madded[0:2, 0:2]
    M01 = Madded[0:2, 2:4]
    M10 = Madded[2:4, 0:2]
    M11 = Madded[2:4, 2:4]

    Ms_list = (M00, M01, M10, M11)

    if distance_type.lower() in ['average', 'average-case', 'ac']:
        cost_function_now = _cost_function_difference_dac
    elif distance_type.lower() in ['worst', 'worst-case', 'wc']:
        cost_function_now = _cost_function_difference_dwc
    else:
        raise ValueError(f"Unsuported distance type: {distance_type}")

    additional_arguments = (Ms_list,
                            )

    bounds_list = [(-1, 1) for _ in range(3)]

    ball = lambda x: x[0] ** 2 + x[1] ** 2 + x[2] ** 2
    constraints = scopt.NonlinearConstraint(fun=ball, lb=0, ub=1)
    optimizer_inital = 'differential_evolution'

    maxiter_initial = 10 * 100 ** 1

    params_starting = [1 / np.sqrt(3), 1 / np.sqrt(3), 1 / np.sqrt(3)]

    differential_evolution_options_initial = {
        'maxiter': maxiter_initial,
        'popsize': 24,
        'workers': 1,
        'updating': 'deferred',
        # 'constraints':constraints,
        # 'updating': 'immediate',
        'disp': printing
    }
    # print('ok1')

    # print('ok')

    initial_kwargs = {
        'differential_evolution_kwargs': differential_evolution_options_initial}

    # if optimizer_inital.upper() in ['DIFFERENTIAL_EVOLUTION']:
    #     basinhopping = False
    # else:
    #     basinhopping = True

    # qprint("running differential evolution")
    res = scop_wrapper(
        initial_parameters=params_starting,
        target_function=cost_function_now,
        optimizer_name=optimizer_inital,
        # basinhopping=basinhopping,
        additional_arguments=additional_arguments,
        bounds_list=bounds_list,
        differential_evolution_constraints=constraints,
        # options={'maxiter': maxiter_initial},
        **initial_kwargs

        # differential_evolution_kwargs=differential_evolution_options_initial
    )
    best_xs = res.x
    best_funopt = res.fun

    # print(sum([x**2 for x in best_xs]))

    if distance_type.lower() in __average_case_distance_names__:
        distance = 1 / 2 * np.sqrt(-best_funopt)
    elif distance_type.lower() in __worst_case_distance_names__:
        distance = -best_funopt

    return distance


def _optimization_function_dwc(parameters_state,
                               povm_2q,
                               sign
                               # qubit_index=0
                               ):
    ny, nz = parameters_state[0], parameters_state[1]
    # print('here1')
    norm = ny ** 2 + nz ** 2

    # if nx<0 or ny<0:
    # return 1000*

    if norm > 1.0:
        return 100 * norm

    nx = np.sqrt(1 - norm)

    # half_nx, half_ny, half_nz = nx/2, ny/2, nz/2

    # print('here2')
    # print(nx,ny,nz)
    rho = 1 / 2 * np.eye(2, dtype=complex)
    rho[0, 1] = (nx - 1j * ny) / 2
    rho[1, 0] = (nx + 1j * ny) / 2
    rho[0, 0] += nz / 2
    rho[1, 1] -= nz / 2

    # print('hejunia2')

    return -_solve_max_dop_problem_for_fixed_input_state(input_state=rho,
                                                         povm_2q=povm_2q,
                                                         sign=sign)


def _find_maximal_dop_fixed_sign(povm2q,
                                 target_qubit,
                                 sign,
                                 printing=False):
    povm_2q = __swap_or_not(povm_2q_input=povm2q,
                            target_qubit=target_qubit)
    cost_function_now = _optimization_function_dwc
    additional_arguments = (povm_2q,
                            sign
                            )

    bounds_list = [(-1, 1) for _ in range(2)]

    # bounds_list = None
    optimizer_inital = 'differential_evolution'

    maxiter_initial = 10 * 10 ** 1

    params_starting = [0, 0]

    differential_evolution_options_initial = {
        'maxiter': maxiter_initial,
        'popsize': 24,
        'workers': -1,
        'updating': 'deferred',
        # 'updating': 'immediate',
        'disp': printing
    }
    # print('ok1')

    # print('ok')

    initial_kwargs = {
        'differential_evolution_kwargs': differential_evolution_options_initial}

    # if optimizer_inital.upper() in ['DIFFERENTIAL_EVOLUTION']:
    #     basinhopping = False
    # else:
    #     basinhopping = True

    # qprint("running differential evolution")
    res = scop_wrapper(
        initial_parameters=params_starting,
        target_function=cost_function_now,
        optimizer_name=optimizer_inital,
        # basinhopping=basinhopping,
        additional_arguments=additional_arguments,
        bounds_list=bounds_list,
        # options={'maxiter': maxiter_initial},
        **initial_kwargs
        # differential_evolution_kwargs=differential_evolution_options_initial
    )
    # qprint("GOT IT")
    best_xs = res.x
    best_funopt = res.fun
    #
    # qprint("BEST VAL DIFF:",best_funopt)
    #
    # # raise KeyError("UYAY")
    # res = scop_wrapper(initial_parameters=best_xs,
    #                    target_function=cost_function_now,
    #                    optimizer_name=optimizer_final,
    #                    basinhopping=True,
    #                    additional_arguments=additional_arguments,
    #                    bounds_list=bounds_list,
    #                    options={'maxiter': maxiter_final},
    #                    basinhopping_kwargs=basinhopping_options_final
    #                    )
    #
    # xopt = res.x
    # funopt = res.fun
    #
    # qprint("BEST VAL 2:",funopt)

    return best_funopt


def find_maximal_dop(povm2q,
                     target_qubit):
    # print('hejunia4')
    dop_plus = -_find_maximal_dop_fixed_sign(povm2q=povm2q,
                                             target_qubit=target_qubit,
                                             sign=1)
    dop_minus = -_find_maximal_dop_fixed_sign(povm2q=povm2q,
                                              target_qubit=target_qubit,
                                              sign=-1)

    return np.max([dop_plus, dop_minus])


def get_reduced_noise_matrix_2q(stochastic_map2q,
                                target_qubit,
                                spectator_state):
    stochastic_map2q_output = copy.deepcopy(stochastic_map2q)
    if target_qubit == 1:
        stochastic_map2q_output = Const.standard_gates()['SWAP'].real @ stochastic_map2q_output @ \
                                  Const.standard_gates()['SWAP'].real

    # stochastic map convention
    # 00, 01, 10, 11  <- input state
    # 00----------------
    # 01
    # 10----------------
    # 11
    # ^
    # |
    # output state

    # forgot about convention here :-d
    stochastic_map2q_output = stochastic_map2q_output.T
    if spectator_state in [0, '0']:
        p_0_0 = (stochastic_map2q_output[0, 0] + stochastic_map2q_output[0, 1])
        p_0_1 = (stochastic_map2q_output[2, 0] + stochastic_map2q_output[2, 1])

    elif spectator_state in [1, '1']:
        p_0_0 = (stochastic_map2q_output[1, 0] + stochastic_map2q_output[1, 1])
        p_0_1 = (stochastic_map2q_output[3, 0] + stochastic_map2q_output[3, 1])

    else:
        raise ValueError(f"Wrong neighbors state: {spectator_state}")

    p_1_0 = 1 - p_0_0
    p_1_1 = 1 - p_0_1

    return np.array([[p_0_0, p_0_1],
                     [p_1_0, p_1_1]])


def get_correlation_coefficient_classical(stochastic_map_2q_or_povm,
                                          target_qubit,
                                          distance_type):
    if isinstance(stochastic_map_2q_or_povm, list):
        stochastic_map_2q_or_povm = povmtools.get_stochastic_map_from_povm(stochastic_map_2q_or_povm)

    noise_maps = [get_reduced_noise_matrix_2q(stochastic_map2q=stochastic_map_2q_or_povm,
                                              target_qubit=target_qubit,
                                              spectator_state=i) for i in range(2)]

    povm_0, povm_1 = tuple(
        [povmtools.get_povm_from_stochastic_map(stochastic_map=stochastic_map) for stochastic_map
         in noise_maps])

    if distance_type.lower() in __average_case_distance_names__:
        return average_distance_POVMs(povm_0, povm_1)
    elif distance_type.lower() in __worst_case_distance_names__:
        return operational_distance_POVMs(povm_0, povm_1, classical=True)


def find_correlations_coefficients(povm_2q,
                                   distance_type,
                                   classical=False,
                                   direct_optimization=True):
    if classical:
        c_i_j = get_correlation_coefficient_classical(stochastic_map_2q_or_povm=povm_2q,
                                                      target_qubit=0,
                                                      distance_type=distance_type)
        c_j_i = get_correlation_coefficient_classical(stochastic_map_2q_or_povm=povm_2q,
                                                      target_qubit=1,
                                                      distance_type=distance_type)




    else:
        if distance_type.lower() in __average_case_distance_names__:

            if direct_optimization:
                c_i_j = direct_optimization_difference(povm2q=povm_2q,
                                                       target_qubit=0,
                                                       distance_type='ac')
                c_j_i = direct_optimization_difference(povm2q=povm_2q,
                                                       target_qubit=1,
                                                       distance_type='ac')
            else:

                c_i_j, c_j_i = calculate_correlations_coefficients_average_case(povm_2q=povm_2q)

        elif distance_type.lower() in __worst_case_distance_names__:

            if direct_optimization:
                c_i_j = direct_optimization_difference(povm2q=povm_2q,
                                                       target_qubit=0,
                                                       distance_type='wc')
                c_j_i = direct_optimization_difference(povm2q=povm_2q,
                                                       target_qubit=1,
                                                       distance_type='wc')

            else:
                c_i_j = find_maximal_dop(povm2q=povm_2q,
                                         target_qubit=0)
                c_j_i = find_maximal_dop(povm2q=povm_2q,
                                         target_qubit=1)

        else:
            raise ValueError(f"Distance type '{distance_type}' incorrect.")

    return c_i_j, c_j_i



from collections import Counter
from typing import Optional, List, Dict, Union


#ORGANIZE - this has to be globally accessible list (MO)
__valid_SDK_names__ = ['QISKIT',
                       # 'PYQUIL',
                       'PYQUIL-FOR-AZURE-QUANTUM']


class SDKHandlerBase:
    def __init__(self,
                 SDK_name: str,
                 qubit_indices: List[int],
                 descriptions_of_circuits: Union[Dict[str, int], List[List[int]]],
                 experiment_name: str,
                 number_of_repetitions: Optional[int] = 1,
                 quantum_register_size: Optional[int] = None,  # MOcomm this variable does not seam to be used in relevant way - refactoring needed
                 classical_register_size: Optional[int] = None,  # MOcomm this variable does not seam to be used in relevant way - refactoring needed
                 add_barriers=True,
                 pyquil_compilation_method='parametric-native',
                 temporary_helper_argument=True,
                 ):

        if SDK_name.upper() not in __valid_SDK_names__:
            raise ValueError('Backend: ' + SDK_name + ' is not supported yet.')

        self._SDK_name = SDK_name
        self._experiment_name = experiment_name
        self._qubit_indices = qubit_indices

        self.pyquil_compilation_method = pyquil_compilation_method

        # TODO FBM: very temporary
        if isinstance(descriptions_of_circuits, list) and temporary_helper_argument:
            self._circuits_labels_list = descriptions_of_circuits
            circuits_labels_strings_list = [''.join([str(symbol) for symbol in symbols_list]) for
                                            symbols_list
                                            in
                                            descriptions_of_circuits]
            descriptions_of_circuits = dict(zip(Counter(circuits_labels_strings_list).keys(),
                                       Counter(circuits_labels_strings_list).values()))
            if number_of_repetitions > 1:
                for key in descriptions_of_circuits.keys():
                    descriptions_of_circuits[key] *= number_of_repetitions

        self._circuit_labels = descriptions_of_circuits
        self._quantum_register_size = quantum_register_size
        self._classical_register_size = classical_register_size
        self._add_barriers = add_barriers

    @staticmethod
    def _add_measurements_qiskit(
            circuit_object,
            qreg,
            creg,
            qubit_indices):
        for qubit_index in range(len(qubit_indices)):
            # print(qubit_indices)
            # print(qreg)
            # print(creg)
            circuit_object.measure(qreg[qubit_indices[qubit_index]], creg[qubit_index])

        return circuit_object

    @staticmethod
    def _add_measurements_pyquil(
            quantum_program,
            qubit_indices,
            classical_register):
        from pyquil.gates import MEASURE
        for qubit_index in range(len(qubit_indices)):
            quantum_program += MEASURE(qubit_indices[qubit_index],
                                       classical_register[qubit_index])
        return quantum_program

    def _add_measurements(self,
                          quantum_circuit,
                          qubit_indices,
                          classical_register,
                          quantum_register=None):

        1
        # TODO FBM: finish this

#
    def __create_circuit_qiskit(self,
                                quantum_register_size: int,
                                classical_register_size: int,
                                # qubit_indices: List[int],
                                # circuit_label_list: List[Union[int, str]],
                                circuit_name: str,
                                # add_measurements: bool
                                ):
#MOcomm -imports next to evaluation, not elegant!
        from qiskit import QuantumCircuit, QuantumRegister, ClassicalRegister
        # create registers
        qreg, creg = QuantumRegister(quantum_register_size), \
                     ClassicalRegister(classical_register_size)
        # print(qreg)
        circuit_object = QuantumCircuit(qreg,
                                        creg,
                                        name=circuit_name)

        return circuit_object, qreg, creg

#MOcomm - what is classical register size below?

    def __create_circuit_pyquil(self,
                                # quantum_register_size: int,
                                classical_register_size: int,
                                qubit_indices: List[int], # not used inside the method, but returned by it
                                # pyquil_compilation_method:str='parametric-native'
                                # circuit_label_list: List[Union[int, str]],
                                # circuit_name: str,
                                # add_measurements: bool
                                ):

        from pyquil import Program

        circuit_object = Program()
        classical_register = circuit_object.declare('ro',
                                                    'BIT',
                                                    classical_register_size)

        return circuit_object, qubit_indices, classical_register

    def __create_circuit_braket(self,
                                # quantum_register_size: int,
                                classical_register_size: int,
                                qubit_indices: List[int],
                                qubi
                                ):
        from braket.circuits import Circuit
        circuit_object = Circuit()
        return circuit_object


    def _create_circuit(self,
                        classical_register_size,
                        qubit_indices=None,
                        circuit_name=None,
                        quantum_register_size=None):
        # TODO JM, FBM: add braket

        if self._SDK_name.upper() in ['QISKIT']:
            if quantum_register_size is None:
                raise ValueError("Please provide quantum register size for qiskit!")
            if circuit_name is None:
                raise ValueError("Please provide circuit name for qiskit!")
            circuit_object, qreg, creg = self.__create_circuit_qiskit(quantum_register_size=quantum_register_size,
                                                                      classical_register_size=classical_register_size,
                                                                      # qubit_indices=qubit_indices,
                                                                      circuit_name=circuit_name)

        elif self._SDK_name.upper() in ['PYQUIL', 'PYQUIL-FOR-AZURE-QUANTUM']:
            if qubit_indices is None:
                raise ValueError("Please provide qubits indices for pyquil!")
            circuit_object, qreg, creg = self.__create_circuit_pyquil(classical_register_size=classical_register_size,
                                                                      qubit_indices=qubit_indices)

        elif self._SDK_name.upper() in ['BRAKET']:
            circuit_object = self.__create_circuit_braket()
            qreg = None
            creg = None

        return circuit_object, qreg, creg

#
#



"""
@author: Jan Tuziemski, Filip Maciejewski, Joanna Majsak, Oskar Słowik, Marcin Kotowski, Katarzyna Kowalczyk-Murynka, Paweł Przewłocki, Piotr Podziemski, Michał Oszmaniec
@contact: filip.b.maciejewski@gmail.com
"""
from typing import Callable, List, Tuple, Optional

import numpy as np
from scipy import optimize as scoptimize
from scipy.optimize import LinearConstraint


from qrem.common.printer import qprint

def wrapper_scipy_optimize(
        target_function: Callable,
        optimizer_name: str,
        initial_parameters: Optional[List[float]]=None,
        basinhopping=False,
        basinhopping_kwargs={},
        differential_evolution_kwargs={},
        additional_arguments: Optional[Tuple] = (),
        bounds_list: Optional[List[Tuple[float,float]]] = None,
        options: Optional[dict] = None,
        differential_evolution_constraints = ()
):


    if optimizer_name.upper() not in ['DIFFERENTIAL_EVOLUTION'] and initial_parameters is None:
        raise ValueError(f"Optimizer '{optimizer_name}' requires initial parameters guess!")


    if optimizer_name.upper() in ['DIFFERENTIAL_EVOLUTION'] and basinhopping:
        raise ValueError("Chosen optimizer 'differential evolution' does not support basinhopping,"
                         " but it was set to True.")


    number_of_parameters = len(initial_parameters)
    if bounds_list is not None:
        if optimizer_name in ['COBYLA', 'SLSQP']:
            constraints_list = []
            for lower, upper in bounds_list:
                def _ineq_fuction_now(x):
                    return (lower - x) * (x - upper)

                constraints_list.append({'type': 'ineq',
                                         'fun': _ineq_fuction_now})
            bounds_list = None

        elif optimizer_name == 'trust-constr':
            constraints_list = LinearConstraint(np.eye(number_of_parameters),
                                                lb=[tup[0] for tup in bounds_list],
                                                ub=[tup[1] for tup in bounds_list])

        else:
            bounds_list = bounds_list
            constraints_list = ()
    else:
        constraints_list = ()

    if basinhopping:

        res = scoptimize.basinhopping(func=target_function,
                                      x0=initial_parameters,
                                      **basinhopping_kwargs,
                                      minimizer_kwargs={'method': optimizer_name,
                                                        'options': options,
                                                        'args': additional_arguments,
                                                        'constraints':constraints_list,
                                                        'bounds':bounds_list}
                                      # minimizer_kwargs=dict,
                                      )

    else:
        if optimizer_name.upper() == 'DIFFERENTIAL_EVOLUTION':
            # print('printing hej')
            res = scoptimize.differential_evolution(func=target_function,
                                                    bounds=bounds_list,
                                                    args=additional_arguments,
                                                    constraints=differential_evolution_constraints,
                                                    **differential_evolution_kwargs
                                                    )
            # print('printing hej2')

        else:

            res = scoptimize.minimize(fun=target_function,
                                      x0=initial_parameters,
                                      args=additional_arguments,
                                      method="trust-constr",
                                      bounds=bounds_list,
                                      constraints=constraints_list,
                                      options=options
                                      )

    return res


def fun_test(xs):
    return 2 * xs[0] + 3 * xs[1]


if __name__ == '__main__':
    params_starting = [1, 1]

    number_of_parameters = len(params_starting)

    bounds_list = [(0, 1) for _ in range(number_of_parameters)]

    optimizer_name = 'Nelder-Mead'
    options = {'maxiter': 10 ** 7}
    res = wrapper_scipy_optimize(target_function=fun_test,
                                 initial_parameters=params_starting,
                                 bounds_list=bounds_list,
                                 optimizer_name=optimizer_name,
                                 options=options)

    optimal_parameters = res.x
    optimal_value = res.fun

    qprint('Optimal parameters:', optimal_parameters)
    qprint('Optimal value:', optimal_value)

    res = wrapper_scipy_optimize(target_function=fun_test,
                                 initial_parameters=optimal_parameters,
                                 bounds_list=bounds_list,
                                 optimizer_name=optimizer_name)
    print()
    optimal_parameters = res.x
    optimal_value = res.fun

    qprint('Optimal parameters:', optimal_parameters)
    qprint('Optimal value:', optimal_value)




from typing import Optional, Dict, List, Tuple

import numpy as np
from tqdm import tqdm

from qrem.noise_characterization.base_classes.marginals_analyzer_base import MarginalsAnalyzerBase
from qrem.functions_qrem import povmtools as pt

from qrem.common import probability
from qrem.common.printer import qprint #,qprint_array

class MarginalsCorrector(MarginalsAnalyzerBase):
    """
    This is the main class that uses correction data to reduce noise on the level of marginals_dictionary.
    Main functionalities are to correct marginal distributions of some experiments.

    NOTE: please see parent class MarginalsAnalyzerBase for conventions of experimental results storage

    """

    def __init__(self,
                 experimental_results_dictionary: Dict[str, Dict[str, int]],

                 correction_data_dictionary: dict,
                 bitstrings_right_to_left: Optional[bool] = False,
                 marginals_dictionary: Optional[Dict[str, Dict[Tuple[int], np.ndarray]]] = None
                 ) -> None:
        """
        :param experimental_results_dictionary: dictionary of results of experiments we wish to correct
        (see class' description for conventions)

        :param bitstrings_right_to_left: specify whether bitstrings
                                    should be read from right to left (when interpreting qubit labels)
        :param correction_data_dictionary: dictionary that contains information needed for noise
                                           mitigation on marginal probability distributions.


        :param marginals_dictionary: in case we pre-computed some marginal distributions
                                     (see class' description for conventions)

        """

        super().__init__(results_dictionary=experimental_results_dictionary,
                         bitstrings_right_to_left=bitstrings_right_to_left,
                         marginals_dictionary=marginals_dictionary
                         )
        if 'noise_matrices' in correction_data_dictionary.keys():
            self._noise_matrices = correction_data_dictionary['noise_matrices']
        else:
            qprint('No noise matrices provided!','','red')
        # self._correction_matrices = correction_data_dictionary['correction_matrices']
        if 'correction_matrices' in correction_data_dictionary.keys():
            self._correction_matrices = correction_data_dictionary['correction_matrices']
            # print('hejka')
        else:
            qprint('No correction matrices provided!','','red')



        self._correction_indices = correction_data_dictionary['correction_indices']

        self._corrected_marginals = {}

    @property
    def correction_indices(self) -> Dict[str, str]:
        return self._correction_indices

    @correction_indices.setter
    def correction_indices(self, correction_indices) -> None:
        self._correction_indices = correction_indices

    @property
    def corrected_marginals(self) -> Dict[str, Dict[str, np.ndarray]]:
        return self._corrected_marginals

    def compute_marginals_of_marginals(self,
                                       keys_of_interest: List[str],
                                       corrected=True) -> \
            Dict[str, Dict[str, np.ndarray]]:
        """From dictionary of marginals_dictionary take only those which are in "marginals_labels_hamiltonian".
        Furthermore, for all keys, calculate also two-qubit and single-qubit marginals_dictionary for qubits
        inside those marginals_dictionary.

        :param keys_of_interest: list of strings representing qubit indices, e.g., 'q1q3q15'

        :return: marginals_of_interest : dictionary with marginal distributions for marginals_labels_hamiltonian
        """
        # self._corrected_marginals = dict(self._corrected_marginals)
        # self._marginals_dictionary = dict(self._marginals_dictionary)

        if corrected:
            marginals_dictionary = self._corrected_marginals
        else:
            marginals_dictionary = self._marginals_dictionary

        marginals_of_interest = {}

        for key in marginals_dictionary.keys():
            distribution_now = marginals_dictionary[key]
            if key in keys_of_interest:
                marginals_of_interest[key] = distribution_now

            qubits_here = key
            enumerated_qubits = dict(enumerate(key))
            rev_map = {}
            for kkk, vvv in enumerated_qubits.items():
                rev_map[vvv] = kkk

            
            #Here is one source of a problem

            # print(qubits_here)
            for qi in qubits_here:
                if (qi,) in keys_of_interest:# and not marginals_dictionary.keys():
                    marginals_of_interest[(qi,)] = \
                        self.get_marginal_from_probability_distribution(
                            distribution_now, [rev_map[qi]])

            for qi in qubits_here:
                for qj in qubits_here:
                    if (qi, qj) in keys_of_interest:# and not marginals_dictionary.keys():
                        marginals_of_interest[(qi, qj)] = self.get_marginal_from_probability_distribution(
                            distribution_now, sorted([rev_map[qi], rev_map[qj]]))

        if corrected:
            self._corrected_marginals = {**marginals_dictionary,**marginals_of_interest}

            #return marginals_of_interest
            return self._corrected_marginals
        else:
            self._marginals_dictionary = {**marginals_dictionary,**marginals_of_interest}

            return self._marginals_dictionary



    @staticmethod
    def correct_distribution_T_matrix(distribution: np.ndarray,
                                      correction_matrix: np.ndarray,
                                      ensure_physicality=True,
                                      norm_minimization='l2'):
        """Correct probability distribution using multiplication via inverse of noise matrix.
        See Refs. [0], [0.5].

        :param distribution: noisy distribution
        :param correction_matrix: correction matrix (inverse of noise matrix)
        :param ensure_physicality: if True, then after correcting marginal distribution it ensures that
                                    resulting vector has elements from [0,1]

        :return: array representing corrected distribution

        #TODO FBM: add option to return mitigation errors
        """

        # Consistent formatting

        #JT: in the definition distribution is specified as np.ndarray, do we need the ib below?
        if isinstance(distribution, list):
            d = len(distribution)
        elif isinstance(distribution, np.ndarray):
            d = distribution.shape[0]
        else:
            raise TypeError("Wrong distribution format")

        distribution = np.array(distribution).reshape(d, 1)

        # correct distribution using inverse of noise matrix
        quasi_distribution = correction_matrix.dot(distribution)

        # do we want to ensure resulting distribution is physical?
        if ensure_physicality:
            # if yes, check if it is physical
            if probability.is_valid_probability_vector(quasi_distribution):
                # if it is physical, no further action is required
                return quasi_distribution
            else:
                # if it is not physical, find closest physical one
                if norm_minimization.lower() in ['l2']:
                    return np.array(pt.find_closest_prob_vector_l2(quasi_distribution)).reshape(d, 1)
                elif norm_minimization.lower() in ['l1']:
                    return np.array(pt.find_closest_prob_vector_l1(quasi_distribution)).reshape(d, 1)
        else:
            # if we don't care about physicality, we don't do anything and just return vector
            return quasi_distribution

    @staticmethod
    def correct_distribution_IBU(estimated_distribution: np.ndarray,
                                 noise_matrix: np.ndarray,
                                 iterations_number: Optional[int] = 10,
                                 prior: Optional[np.ndarray] = None):
        """Correct probability distribution using Iterative Bayesian Unfolding (IBU)
        See Ref. [1]

        :param estimated_distribution: noisy distribution (to be corrected)
        :param noise_matrix: noise matrix
        :param iterations_number: number of iterations in unfolding
        :param prior: ansatz for ideal distribution, default is uniform

        :return: array representing corrected distribution
        """

        # Consistent formatting
        if isinstance(estimated_distribution, list):
            d = len(estimated_distribution)
        elif isinstance(estimated_distribution, np.ndarray):
            d = estimated_distribution.shape[0]
        else:
            raise TypeError("Wrong distribution format")

        # default iterations number
        if iterations_number is None:
            iterations_number = 10

        # If no prior is provided, we take uniform distribution
        if prior is None:
            prior = np.full((d, 1), 1 / d, dtype=float)
        elif isinstance(prior,str):
            if prior in ['uniform']:
                prior = np.full((d, 1), 1 / d, dtype=float)

            # max_index = np.argmax(estimated_distribution)
            # prior = np.ones((d,1))
            # prior[max_index] = 10
            # prior/=np.sum(prior)


        # raise KeyboardInterrupt
        # initialize distribution for recursive loop
        distribution_previous = prior
        # print(noise_matrix)
        # go over iterations
        for iteration_index in range(iterations_number):
            distribution_new = np.zeros((d, 1),
                                        dtype=float)

            # go over measurement outcomes
            for outcome_index in range(d):
                # calculating IBU estimate
                distribution_new[outcome_index] = sum(
                    [noise_matrix[j, outcome_index] * distribution_previous[outcome_index] *
                     estimated_distribution[j] / sum(
                        [noise_matrix[j, k] * distribution_previous[k] for k in range(d)])
                     for j in
                     range(d)])
            # update distribution for recursion
            distribution_previous = distribution_new

        return distribution_previous

    # @staticmethod
    def correct_distribution_hybrid_T_IBU(self,
                                          estimated_distribution,
                                          noise_matrix,
                                          correction_matrix: Optional[np.ndarray] = None,
                                          unphysicality_threshold: Optional[float] = 0.0,
                                          iterations_number: Optional[int] = None,
                                          prior: Optional[np.ndarray] = None):
        """
        Correct distribution using method_name that is hybrid of T_matrix correction (see Refs. [0,0.5])
        and Iterative Bayesian Unfolding (IBU) (see Ref. [1]).

        Algorithm goes like this:
        - Correct distribution using inverse of noise matrix.
        - If distribution is physical (i_index.e., elements are from (0,1)), return it.
        - Otherwise, perform IBU.
        - [Optional] if parameter "unphysicality_threshold" is provided,
           then it projects the result of "T_matrix" correction onto probability simplex
           and if Total Variation Distance between this projected vector
           and original unphysical one is belowthis cluster_threshold, then it returns the projection.
           Otherwise, it performs IBU.

        :param estimated_distribution: noisy distribution (to be corrected)
        :param noise_matrix: noise matrix
        :param correction_matrix: correction matrix (inverse of noise matrix)
                                 if not provided, it is calculated from noise_matrix
        :param unphysicality_threshold: cluster_threshold to decide whether unphysical "T_matrix"
                                        correction is acceptable. See description of the function
        :param iterations_number: number of iterations in IBU
        :param prior: ansatz for ideal distribution in IBU, default is uniform

        """

        if isinstance(estimated_distribution, list):
            d = len(estimated_distribution)
        elif isinstance(estimated_distribution, np.ndarray):
            d = estimated_distribution.shape[0]
        else:
            raise TypeError("Wrong distribution format")

        if correction_matrix is None:
            correction_matrix = np.linalg.inv(noise_matrix)

        distribution = np.array(estimated_distribution).reshape(d, 1)

        quasi_distribution = correction_matrix.dot(distribution)

        is_physical = probability.is_valid_probability_vector(quasi_distribution)

        if is_physical:
            return quasi_distribution
        else:
            if unphysicality_threshold == 0:
                closest_physical = pt.find_closest_prob_vector_l2(quasi_distribution)
                return self.correct_distribution_IBU(distribution,
                                                     noise_matrix,
                                                     iterations_number,
                                                     prior=closest_physical)
            else:
                closest_physical = pt.find_closest_prob_vector_l2(quasi_distribution)

                distance_now = 1 / 2 * np.linalg.norm(quasi_distribution - closest_physical, ord=1)
                # distance_now = 1 / 2 * np.linalg.norm(quasi_distribution - closest_physical, ord=1)



                #TODO FBM: original
                if distance_now <= unphysicality_threshold:
                    return closest_physical

                #
                # if distance_now >= unphysicality_threshold:
                #     return closest_physical

                else:
                    return self.correct_distribution_IBU(distribution,
                                                         noise_matrix,
                                                         iterations_number,
                                                         prior=closest_physical)

    def correct_marginals(self,
                          marginals_dictionary: Optional[Dict[str, np.ndarray]] = None,
                          method='lambda_inverse',
                          method_kwargs=None) -> Dict[str, Dict[str, np.ndarray]]:

        """Return dictionary of corrected marignal distributions
        :param marginals_dictionary: dictionary of (noisy) marginal distributions
        :param method: method_name to be used for correction of marginal probability distributions

        possible values:
        - 'T_matrix' - uses inverse of noise matrix as correction (see Refs. [0,0.5])
        - 'IBU' - uses Iterative Bayesian Unfolding (see Ref. [1])
        - 'hybrid_T_IBU' - uses hybrid method_name between 'T_matrix' and 'IBU',
                          see description of self.correct_distribution_hybrid_T_IBU

        :param method_kwargs:  keyword arguments passed to function using chosen method_name.
                               See description of specific functions_qrem.

        :return: corrected_marginals : dictionary of corrected marginal distributions
        """

        # TODO FBM: make it consistent with conventions used in parent class
        if marginals_dictionary is None:
            marginals_dictionary = self._marginals_dictionary

        corrected_marginals = {}

        for key in marginals_dictionary.keys():

            noisy_marginal_now = marginals_dictionary[key]

            # print(key)
            # qprint_array(self._correction_matrices[key])
            # qprint_array(noisy_marginal_now)

            if method.lower() in ['lambda_inverse',
                                  # 'linear_inversion',
                                  't_matrix']:
                if method_kwargs is None:
                    method_kwargs = {'ensure_physicality': True,
                                     'norm_minimization':'l2'}
                else:
                    if 'ensure_physicality' not in method_kwargs.keys():
                        method_kwargs['ensure_physicality'] = True

                    # for unwanted_key in

                corrected_marginal_now = self.correct_distribution_T_matrix(noisy_marginal_now,
                                                                            self._correction_matrices[
                                                                                key],
                                                                            **method_kwargs)
            elif method.lower() in  ['ibu']:
                if method_kwargs is None:
                    method_kwargs = {'iterations_number': 2,
                                     'prior': None}
                else:
                    if 'iterations_number' not in method_kwargs.keys():
                        method_kwargs['iterations_number'] = 2
                    if 'prior' not in method_kwargs.keys():
                        method_kwargs['prior']=None


                #TODO FBM: some problems with KEY in self._noise_matrices
                corrected_marginal_now = self.correct_distribution_IBU(noisy_marginal_now,
                                                                       self._noise_matrices[
                                                                           key]['averaged'],
                                                                       **method_kwargs)
            elif method.lower() in ['hybrid_lambda_ibu']:
                if method_kwargs is None:
                    method_kwargs = {'unphysicality_threshold': 0.0,
                                     'iterations_number': 2,
                                     'prior': None}
                else:
                    if 'iterations_number' not in method_kwargs.keys():
                        method_kwargs['iterations_number'] = 2
                    if 'prior' not in method_kwargs.keys():
                        method_kwargs['prior'] = None
                    if 'unphysicality_threshold' not in method_kwargs.keys():
                        method_kwargs['unphysicality_threshold'] = 0.0

                corrected_marginal_now = self.correct_distribution_hybrid_T_IBU(noisy_marginal_now,
                                                                                self._noise_matrices[
                                                                                    key]['averaged'],
                                                                                self._correction_matrices[
                                                                                    key],
                                                                                **method_kwargs)
            else:
                raise ValueError('Wrong method_name name')

            corrected_marginals[key] = corrected_marginal_now

        self._corrected_marginals = corrected_marginals
        return corrected_marginals



"""
@author: Jan Tuziemski, Filip Maciejewski, Joanna Majsak, Oskar Słowik, Marcin Kotowski, Katarzyna Kowalczyk-Murynka, Paweł Przewłocki, Piotr Podziemski, Michał Oszmaniec
@contact: filip.b.maciejewski@gmail.com
"""

import time
from collections import Counter, defaultdict
from typing import Dict, Optional, Tuple, List, Union

import numpy as np

from qrem.functions_qrem import functions_data_analysis as fda, ancillary_functions as anf, \
    quantum_ancillary_functions as quanf
from qrem.functions_qrem.functions_data_analysis import _sort_clusters_division
from qrem.noise_mitigation.base_classes.EnergyEstimatorBase import EnergyEstimatorBase
from qrem.noise_model_generation.CTMP import functions_CTMP as fun_CTMP
from qrem.noise_simulation.CTMP import functions_CTMP_conditional as CTMP_cond, \
    functions_CTMP_simple as CTMP_simp

from qrem.common.printer import qprint
from qrem.common import convert

class EnergyCorrectorLocalCTMP(EnergyEstimatorBase):
    """
    This is class for implementing quasiprobability-like error mitigation based on CTMP model [#TODO FBM: add ref].

    """

    # TODO FBM: test whether all fo this works in toy examples

    def __init__(self,
                 results_dictionary: Dict[str, Dict[str, int]],
                 hamiltonian_dictionaries: Dict[str, Dict[Tuple[int], float]],
                 CTMP_rates_dictionary: Dict[Tuple[int], Dict[str, float]],
                 clusters_list: List[Tuple[int]],
                 local_noise_strengths: Optional[Dict[Tuple[int], float]] = None,
                 local_noise_matrices: Optional[Dict[Tuple[int], np.ndarray]] = None

                 ) -> None:
        """
        :param results_dictionary: see parent class description
        :param hamiltonian_dictionaries: see parent class description
        :param CTMP_rates_dictionary: dictionary of error rates
        """

        super().__init__(results_dictionary=results_dictionary,
                         hamiltonian_dictionaries=hamiltonian_dictionaries)

        self._CTMP_rates_dictionary = CTMP_rates_dictionary
        self._energies_dictionary_corrected = {}
        self._default_max_number_of_samples = 10 ** 5
        self._clusters_list = clusters_list

        local_rates_dictionaries = fun_CTMP._convert_to_local_noise_rates_dictionary(
            clusters_list=clusters_list,
            CTMP_rates_dictionary=CTMP_rates_dictionary)

        if local_noise_strengths is None:
            local_noise_strengths = fun_CTMP.calculate_noise_strengths_local(
                local_rates_dictionaries=local_rates_dictionaries,
                method='bruteforce')
        if local_noise_matrices is None:
            local_noise_matrices = fun_CTMP.get_local_noise_matrices(
                local_rates_dictionaries=local_rates_dictionaries,
                local_noise_strengths=local_noise_strengths)
        self._local_noise_strengths = local_noise_strengths
        self._local_noise_matrices = local_noise_matrices

        matrix_powers_dictionary = {cluster: {} for cluster in clusters_list}
        for cluster in clusters_list:
            for matrix_power in range(11):
                matrix_powers_dictionary[cluster][matrix_power] = np.linalg.matrix_power(
                    local_noise_matrices[cluster], matrix_power)

        self._matrix_powers_dictionary = matrix_powers_dictionary

        qubits_to_clusters_map = {}
        for cluster in clusters_list:
            for qi in cluster:
                qubits_to_clusters_map[(qi,)] = cluster

        self._qubits_to_clusters_map = qubits_to_clusters_map

        self.__CTMP_generator_matrices = fun_CTMP.get_generators_dictionary_from_rates(
            rates_dictionary=CTMP_rates_dictionary)

    @property
    def CTMP_rates_dictionary(self) -> Dict[Tuple[int], Dict[str, float]]:
        return self._CTMP_rates_dictionary

    @CTMP_rates_dictionary.setter
    def CTMP_rates_dictionary(self, CTMP_rates_dictionary: Dict[Tuple[int], Dict[str, float]]) -> None:
        self._CTMP_rates_dictionary = CTMP_rates_dictionary

    def estimate_corrected_energy_CTMP_local_separable(self,
                                                       experiment_key: str,
                                                       number_of_samples_dictionary: Optional[
                                                           Union[int, Dict[Tuple[int], int]]] = None,
                                                       target_estimation_accuracy: Optional[
                                                           float] = 0.01,
                                                       ):


        local_noise_strengths = self._local_noise_strengths
        clusters_list_global = self._clusters_list
        qubits_to_clusters_map = self._qubits_to_clusters_map
        matrix_powers_dictionary = self._matrix_powers_dictionary
        local_noise_matrices = self._local_noise_matrices

        # import time
        t0 = time.time()
        if number_of_samples_dictionary is None:
            number_of_samples_dictionary = {}
            for i in range(len(clusters_list_global)):
                cluster_i = clusters_list_global[i]
                samples_i = fun_CTMP.get_theoretical_number_of_samples_CTMP(
                    noise_strength_CTMP=local_noise_strengths[cluster_i],
                    target_estimation_accuracy=target_estimation_accuracy)
                number_of_samples_dictionary[(cluster_i,)] = int(
                    np.min([samples_i, self._default_max_number_of_samples]))
                for j in range(i + 1, len(clusters_list_global)):
                    cluster_j = clusters_list_global[j]

                    sorted_tuple = _sort_clusters_division([cluster_i, cluster_j])
                    samples_ij = fun_CTMP.get_theoretical_number_of_samples_CTMP(
                        noise_strength_CTMP=local_noise_strengths[cluster_i] + local_noise_strengths[
                            cluster_j],
                        target_estimation_accuracy=target_estimation_accuracy)
                    number_of_samples_dictionary[sorted_tuple] = int(
                        np.min([samples_ij, self._default_max_number_of_samples]))
            # print(number_of_samples_dictionary)


        elif isinstance(number_of_samples_dictionary, int):
            number_of_samples_dict = {}
            for i in range(len(clusters_list_global)):
                cluster_i = clusters_list_global[i]
                number_of_samples_dict[(cluster_i,)] = number_of_samples_dictionary
                for j in range(i + 1, len(clusters_list_global)):
                    cluster_j = clusters_list_global[j]
                    sorted_tuple = _sort_clusters_division([cluster_i, cluster_j])
                    number_of_samples_dict[sorted_tuple] = number_of_samples_dictionary
            number_of_samples_dictionary = number_of_samples_dict

        # for key, value in number_of_samples_dictionary.items():
        #     print(key,value)
        # raise KeyboardInterrupt
        # print(number_of_samples_dictionary)

        experimental_results = self._results_dictionary[experiment_key]
        hamiltonian_weights = self._hamiltonian_dictionaries[experiment_key]

        # global_output_state_template = ['9'] * self._number_of_qubits

        experimental_outcomes = list(experimental_results.keys())
        probability_distribution_over_outcomes = np.array(list(experimental_results.values()),
                                                          dtype=float)
        probability_distribution_over_outcomes /= np.sum(probability_distribution_over_outcomes)

        local_normalizations_dict = {cluster: np.exp(2 * local_noise_strengths[cluster]) for cluster in
                                     clusters_list_global}

        local_registers_dictionary = {len(cluster):
                                          [tuple(convert.integer_to_bitstring(index_outcome, len(cluster)))
                                           for index_outcome in range(int(2 ** len(cluster)))]
                                      for cluster in clusters_list_global}

        local_registers_dictionary_REVERSED = {len(cluster):
                                                   {convert.integer_to_bitstring(index_outcome,
                                                                             len(cluster)):
                                                        index_outcome
                                                    for index_outcome in range(int(2 ** len(cluster)))}
                                               for cluster in clusters_list_global}

        local_bitstrings_experimental_dict = {bitstring:
                                                  {cluster: ''.join([bitstring[qi] for qi in cluster])
                                                   for cluster in clusters_list_global}
                                              for bitstring in experimental_outcomes}

        local_sets_dictionary = {}

        all_outcomes_samples_uniform = {}
        for i in range(len(clusters_list_global)):

            cluster = clusters_list_global[i]
            number_of_samples_i = number_of_samples_dictionary[(cluster,)]
            all_outcomes_samples_uniform[(cluster,)] = np.random.multinomial(n=number_of_samples_i,
                                                                             pvals=probability_distribution_over_outcomes)

            local_sets_dictionary[(cluster,)] = []

            for j in range(i + 1, len(clusters_list_global)):
                cluster2 = clusters_list_global[j]
                sorted_tuple = _sort_clusters_division([cluster, cluster2])

                number_of_samples_i_j = number_of_samples_dictionary[sorted_tuple]

                # sorted_tuple = _sort_clusters_division([cluster, cluster2])
                all_outcomes_samples_uniform[sorted_tuple] = np.random.multinomial(
                    n=number_of_samples_i_j,
                    pvals=probability_distribution_over_outcomes)

                local_sets_dictionary[sorted_tuple] = []

        for pair in hamiltonian_weights.keys():
            qi = pair[0]
            cluster_i = qubits_to_clusters_map[(qi,)]
            if len(pair) == 1:
                continue
            qj = pair[1]
            cluster_j = qubits_to_clusters_map[(qj,)]

            if cluster_i == cluster_j:
                local_sets_dictionary[(cluster_i,)].append(pair)
            else:
                sorted_tuple = _sort_clusters_division([cluster_i, cluster_j])
                local_sets_dictionary[sorted_tuple].append(pair)
        # t1=time.time()
        # qprint("PREPROCESSING:",t1-t0)
        # raise KeyboardInterrupt

        # t1=  time.time()
        rng = np.random.default_rng()
        energy_estimator = 0
        t_sampling, t_formatting, t_energy, t_filling = 0, 0, 0, 0
        for clusters_tuple, pairs_list in local_sets_dictionary.items():
            qubits_in_clusters_tuple = []
            for cluster in clusters_tuple:
                for qi in cluster:
                    qubits_in_clusters_tuple.append(qi)

            counts_dictionary_now = {}
            samples_uniform_outcomes = all_outcomes_samples_uniform[clusters_tuple]
            outcomes_range = range(len(samples_uniform_outcomes))

            # TODO FBM: temporary

            for index_outcome in outcomes_range:
                number_of_samples_from_outcome = samples_uniform_outcomes[index_outcome]
                bitstring_now = experimental_outcomes[index_outcome]

                samples_pairs = []
                for cluster in clusters_tuple:
                    local_input_state_index = local_registers_dictionary_REVERSED[len(cluster)][
                        local_bitstrings_experimental_dict[bitstring_now][cluster]]

                    noise_strength_now = local_noise_strengths[cluster]

                    t_sampling_start0 = time.time()
                    local_matrix_powers = rng.poisson(lam=noise_strength_now,
                                                      size=number_of_samples_from_outcome)
                    local_matrix_powers_dict = defaultdict(float)
                    t_sampling += time.time() - t_sampling_start0
                    for power in local_matrix_powers:
                        local_matrix_powers_dict[power] += 1
                    local_samples = []
                    for local_matrix_power, number_of_outcomes_per_power \
                            in local_matrix_powers_dict.items():
                        if (local_matrix_power % 2) == 0:
                            local_sign = 1
                        else:
                            local_sign = -1
                        try:
                            local_matrix_now = matrix_powers_dictionary[cluster][local_matrix_power]
                        except(KeyError):
                            matrix_powers_dictionary[cluster][
                                local_matrix_power] = np.linalg.matrix_power(
                                local_noise_matrices[cluster], local_matrix_power)
                            local_matrix_now = matrix_powers_dictionary[cluster][local_matrix_power]

                        local_distro_now = local_matrix_now[:, local_input_state_index]
                        t_sampling_start1 = time.time()
                        local_output_states = rng.multinomial(n=number_of_outcomes_per_power,
                                                              pvals=local_distro_now)
                        t_sampling += time.time() - t_sampling_start1

                        t_filling0 = time.time()
                        for further_index in range(len(local_output_states)):
                            local_samples += [(local_sign,
                                               local_registers_dictionary[len(cluster)][further_index]
                                               )
                                              ] * local_output_states[further_index]
                        t_filling += time.time() - t_filling0

                    # rng = np.random.default_rng()
                    t_sampling_start2 = time.time()
                    rng.shuffle(local_samples)
                    # all_samples[cluster] = local_samples
                    samples_pairs.append(local_samples)
                    t_sampling += time.time() - t_sampling_start2
                #
                t3 = time.time()


                if len(clusters_tuple) == 1:
                    samples_pairs = list(samples_pairs[0])
                    counting = Counter(samples_pairs)
                    for tuple_outcome, amount_of_ticks in counting.items():
                        global_output_state = ['9'] * self._number_of_qubits
                        cluster_local = clusters_tuple[0]
                        for qubit_index in range(len(cluster_local)):
                            global_output_state[cluster_local[qubit_index]] = \
                                tuple_outcome[1][qubit_index]


                        global_output_state = ''.join(global_output_state)
                        if global_output_state not in counts_dictionary_now.keys():
                            counts_dictionary_now[global_output_state] = defaultdict(float)

                        counts_dictionary_now[global_output_state][
                            (tuple_outcome[0],)] += amount_of_ticks

                else:
                    samples_pairs = list(zip(*samples_pairs))
                    # print(samples_pairs[0])

                    # raise KeyboardInterrupt
                    counting = Counter(samples_pairs)
                    for tuple_outcome, amount_of_ticks in counting.items():
                        global_output_state = ['9'] * self._number_of_qubits
                        for clust_index in range(len(clusters_tuple)):
                            cluster_local = clusters_tuple[clust_index]
                            for qubit_index in range(len(cluster_local)):
                                global_output_state[cluster_local[qubit_index]] = \
                                    tuple_outcome[clust_index][1][qubit_index]


                        global_output_state = ''.join(global_output_state)
                        if global_output_state not in counts_dictionary_now.keys():
                            counts_dictionary_now[global_output_state] = defaultdict(float)

                        counts_dictionary_now[global_output_state][
                            tuple([x[0] for x in tuple_outcome])] += amount_of_ticks

                t4 = time.time()
                t_formatting += t4 - t3
                #
                # t5= time.time()
                # for i in range(number_of_samples_from_outcome):
                #     global_output_state = ['9']*self._number_of_qubits
                #     local_signs = [all_samples[cluster][i][0] for cluster in clusters_tuple]
                #     for cluster in clusters_tuple:
                #         local_output_state = all_samples[cluster][i][1]
                #         for qubit_index in range(len(cluster)):
                #             global_output_state[cluster[qubit_index]] = local_output_state[qubit_index]
                #
                #     global_output_state = ''.join(global_output_state)
                #     if global_output_state not in counts_dictionary_now.keys():
                #         counts_dictionary_now[global_output_state] = defaultdict(float)
                #
                #     counts_dictionary_now[global_output_state][tuple(local_signs)] += 1
                # t6=time.time()
                # t_formatting+=t6-t5

            t5 = time.time()
            # t23+=t3-t2
            #TODO FBM: Make sure it does take into account every possible case
            local_weights_now = {}
            for pair in pairs_list:
                if pair not in hamiltonian_weights.keys():
                    continue
                # weight_now = hamiltonian_weights[pair]
                local_weights_now[pair] = hamiltonian_weights[pair]
            if len(clusters_tuple) == 1:
                for cluster in clusters_tuple:
                    for qi in cluster:
                        if (qi,) not in hamiltonian_weights.keys():
                            continue
                        local_weights_now[(qi,)] = hamiltonian_weights[(qi,)]

            local_energy = 0
            for global_output_state, signs_dict in counts_dictionary_now.items():
                for local_signs, value in signs_dict.items():
                    additional_multipliers_now = {}
                    # TODO FBM: careful here
                    for pair in local_weights_now.keys():
                        qi = pair[0]
                        cluster_i = qubits_to_clusters_map[(qi,)]
                        cluster_index_i = 0
                        sign_i = local_signs[cluster_index_i]

                        local_sign = sign_i
                        local_normalization_now = local_normalizations_dict[cluster_i]

                        # additional_multipliers_now[(qi,)] = local_sign*local_normalization_now*value

                        if len(pair) == 2:
                            qj = pair[1]
                            cluster_j = qubits_to_clusters_map[(qj,)]
                            if cluster_i != cluster_j:
                                cluster_index_j = 1
                                sign_j = local_signs[cluster_index_j]
                                local_sign *= sign_j
                                local_normalization_now *= local_normalizations_dict[cluster_j]

                        # else:
                        #     print('hej',pair)

                        additional_multipliers_now[pair] = local_sign * local_normalization_now * value

                        # for qk in pair:
                        #     additional_multipliers_now[(qk,)]

                    energy_local_now = fda.get_energy_from_bitstring_diagonal_local(
                        bitstring=global_output_state,
                        weights_dictionary=local_weights_now,
                        additional_multipliers=additional_multipliers_now)

                    # energy_local_now *= local_sign * local_normalization_now*value
                    local_energy += energy_local_now

            energy_estimator += local_energy / number_of_samples_dictionary[clusters_tuple]

            t6 = time.time()
            t_energy += t6 - t5

        qprint("(sampling, formatting counts,filling counts, energy estimation):",
                       (t_sampling, t_formatting, t_filling, t_energy))
        energy_estimated = energy_estimator

        self._energies_dictionary_corrected[experiment_key] = energy_estimated

        return energy_estimated




from threading import Thread, Lock
from typing import Optional, Callable, List
import functools as ft

import numpy as np
import numpy.typing as npt

import qrem.common.math as qrem_math



#this function is multithreaded.
#MOVE_TO qrem.common.math
class GeneralTensorCalculator:

    def __init__(self, f:Optional[Callable]=qrem_math.kronecker_product):
        self.f = f
        self.tensor = dict()
        self.lock = Lock()

    # Calculates tensor for given object using function f provided in the constructor.
    def calculate_tensor(self, objects: list) -> dict:
        self.tensor = dict()

        # Get indices combinations
        indices_combinations = self.__get_indices_combinations(objects.copy())

        # Initialize tensor with empty values
        self.__initialize_tensor(indices_combinations)

        # Prepare for multiprocessing
        threads = []

        # Fill tensor dict with values
        for combination in indices_combinations:
            thread = Thread(target=self.count_tensor_value_for_combination, args=(combination, objects,))
            threads.append(thread)
            thread.start()

        for thread in threads:
            thread.join()

        return self.tensor

    # Calculates tensor for given object using function f provided in the constructor, but returns
    # results as a list with increasing order.
    def calculate_tensor_to_increasing_list(self, objects: list) -> list:
        tensor_dict = self.calculate_tensor(objects)
        tensor_list = self.__get_increasing_objects_dict_from_list(tensor_dict)
        return tensor_list

    def __get_increasing_objects_dict_from_list(self, d: dict) -> list:
        objects_list = []
        dict_keys = list(d.keys())

        # This may cause problem if the tensor was somehow calculated for dicts, if it's even possible.
        if type(d[dict_keys[0]]) == dict:
            for k in dict_keys:
                objects_list += self.__get_increasing_objects_dict_from_list(d[k])
        else:
            for k in dict_keys:
                objects_list.append(d[k])

        return objects_list

    # This method has to be public in order to be used in multiprocessing.
    def count_tensor_value_for_combination(self, combination, objects):

        # First count function value for better multiprocessing performance
        f_arguments = []

        # Note that len of objects should be equal to len of current_combination as it contains index value
        # for each object.
        for i in range(len(objects)):
            f_arguments += [objects[i][combination[i]]]

        # Calculate function value now, so that tensor is locked for less time.
        function_value = self.f(f_arguments)

        #(PP) this should be changed to a with self.lock block
        self.lock.acquire()

        try:
            tensor = self.tensor
            for i in range(len(combination) - 1):
                tensor = tensor[combination[i]]
            tensor[combination[-1]] = function_value
        finally:
            self.lock.release()

    # Get combinations of all possible indices variation.
    def __get_indices_combinations(self, objects: list) -> list:
        # Finish if objects list is empty. This should not happen.
        if len(objects) == 0:
            return None

        combinations = []

        # Termination condition
        # If it's the last element of the list of objects create and return list of it's possible indexes
        if len(objects) == 1:
            for i in range(len(objects[0])):
                combinations.append([i])
            return combinations

        # Recursive call
        # If there are still multiple objects take first and make recursive call
        first_object = objects.pop(0)
        other_objects_combination = self.__get_indices_combinations(objects.copy())

        # Then add to combinations sum of other_objects_combinations and popped object indices
        # Note that order is important here
        for i in range(len(first_object)):
            for combination in other_objects_combination:
                combinations.append([i] + combination)

        return combinations

    # Initializing tensor as dict of dict of dict ... with zeros. Initialization for each combination is performed
    # once at a time.
    def __initialize_tensor(self, indices_combination) -> None:
        # Error check.
        if len(indices_combination) == 0:
            return

        for combination in indices_combination:
            self.__initialize_combination_path(self.tensor, combination.copy())

    # Initializing tensor with 0 for given path
    def __initialize_combination_path(self, tensor, combination):
        # Error check
        if len(combination) == 0:
            return

        # Termination condition
        if len(combination) == 1:
            tensor[combination[0]] = 0
            return

        # Recursive call
        initialized_index = combination.pop(0)
        if not tensor.keys().__contains__(initialized_index):
            tensor[initialized_index] = dict()
        self.__initialize_combination_path(tensor[initialized_index], combination)



# -*- coding: utf-8 -*-
"""
The module contains helper functions to work with PICOS.

Created on Wed Dec 10 18:33:34 2014

@author: Peter Wittek
"""
from __future__ import division, print_function
import numpy as np


def solve_with_cvxopt(sdp, solverparameters=None):
    """Helper function to convert the SDP problem to PICOS
    and call CVXOPT solver, and parse the output.

    :param sdp: The SDP relaxation to be solved.
    :type sdp: :class:`ncpol2sdpa.sdp`.
    """
    P = convert_to_picos(sdp)
    P.set_option("solver", "cvxopt")
    P.set_option("verbose", sdp.verbose)
    if solverparameters is not None:
        for key, value in solverparameters.items():
            P.set_option(key, value)
    solution = P.solve()
    x_mat = [np.array(P.get_valued_variable('X'))]
    y_mat = [np.array(P.get_constraint(i).dual)
             for i in range(len(P.constraints))]
    return -solution["cvxopt_sol"]["primal objective"] + \
        sdp.constant_term, \
        -solution["cvxopt_sol"]["dual objective"] + \
        sdp.constant_term, \
        x_mat, y_mat, solution["status"]


def convert_to_picos(sdp, duplicate_moment_matrix=False):
    """Convert an SDP relaxation to a PICOS problem such that the exported
    .dat-s file is extremely sparse, there is not penalty imposed in terms of
    SDP variables or number of constraints. This conversion can be used for
    imposing extra constraints on the moment matrix, such as partial transpose.

    :param sdp: The SDP relaxation to convert.
    :type sdp: :class:`ncpol2sdpa.sdp`.
    :param duplicate_moment_matrix: Optional parameter to add an unconstrained
                                    moment matrix to the problem with the same
                                    structure as the moment matrix with the PSD
                                    constraint.
    :type duplicate_moment_matrix: bool.

    :returns: :class:`picos.Problem`.
    """
    import picos as pic
    import cvxopt as cvx
    P = pic.Problem(verbose=sdp.verbose)
    block_size = sdp.block_struct[0]
    if sdp.F.dtype == np.float64:
        X = P.add_variable('X', (block_size, block_size), vtype="symmetric")
        if duplicate_moment_matrix:
            Y = P.add_variable('Y', (block_size, block_size), vtype="symmetric")
    else:
        X = P.add_variable('X', (block_size, block_size), vtype="hermitian")
        if duplicate_moment_matrix:
            Y = P.add_variable('X', (block_size, block_size), vtype="hermitian")
    row_offset = 0
    theoretical_n_vars = sdp.block_struct[0]**2
    eq_block_start = sdp.constraint_starting_block+sdp._n_inequalities
    block_idx = 0
    while (block_idx < len(sdp.block_struct)):
        block_size = sdp.block_struct[block_idx]
        x, Ix, Jx = [], [], []
        c, Ic, Jc = [], [], []
        for i, row in enumerate(sdp.F.rows[row_offset:row_offset +
                                                     block_size**2]):
            for j, column in enumerate(row):
                if column > 0:
                    x.append(sdp.F.data[row_offset+i][j])
                    Ix.append(i)
                    Jx.append(column-1)
                    i0 = (i//block_size)+(i % block_size)*block_size
                    if i != i0:
                        x.append(sdp.F.data[row_offset+i][j])
                        Ix.append(i0)
                        Jx.append(column-1)

                else:
                    c.append(sdp.F.data[row_offset+i][j])
                    Ic.append(i%block_size)
                    Jc.append(i//block_size)
        permutation = cvx.spmatrix(x, Ix, Jx, (block_size**2,
                                               theoretical_n_vars))
        constant = cvx.spmatrix(c, Ic, Jc, (block_size, block_size))
        if duplicate_moment_matrix:
            constraint = X
        else:
            constraint = X.copy()
        for k in constraint.factors:
            constraint.factors[k] = permutation
        constraint._size = (block_size, block_size)
        if block_idx < eq_block_start:
            P.add_constraint(constant + constraint >> 0)
        else:
            P.add_constraint(constant + constraint == 0)
            row_offset += block_size**2
            block_idx += 1
        if duplicate_moment_matrix and \
                block_size == sdp.block_struct[0]:
            for k in Y.factors:
                Y.factors[k] = permutation
        row_offset += block_size**2
        block_idx += 1
    if len(np.nonzero(sdp.obj_facvar)[0]) > 0:
        x, Ix, Jx = [], [], []
        for k, val in enumerate(sdp.obj_facvar):
            if val != 0:
                x.append(val)
                Ix.append(0)
                Jx.append(k)
        permutation = cvx.spmatrix(x, Ix, Jx)
        objective = X.copy()
        for k in objective.factors:
            objective.factors[k] = permutation
        objective._size = (1, 1)
        P.set_objective('min', objective)
    return P



from typing import Dict, Tuple, List

# checks if transition is available for the given input state
def check_transition(input_state: str, transition: Tuple) -> bool:
    i, j, error_in, _, _ = transition
    if i == j:
        return input_state[i] == error_in
    else:
        return input_state[i] + input_state[j] == error_in

def apply_transition(input_state: str, transition: Tuple) -> str:
    i, j, _, error_out, _ = transition
    state = list(input_state)
    if i == j:
        state[i] = error_out
    else:
        state[i], state[j] = error_out
    return ''.join(state)



from qrem.visualisation.plotting_constants import color_map
from matplotlib import pyplot as plt
from qrem.pipelines.mitigation_functions import compute_mitigation_error_median_mean, compute_mitigation_errors

import numpy as np

from typing import List, Dict

def create_coherence_bound_histogram_data(coherence_bound_dictionary):
    coherence_bound_data_list = []
    for key in coherence_bound_dictionary.keys():
        coherence_bound_data=coherence_bound_dictionary[key][0]
        for data in coherence_bound_data:
            
            coherence_bound_data_list.append(data)
    return coherence_bound_data_list


def create_POVMs_distance_histogram(POVMs_errors_ibm, POVMs_errors_rig, path_to_save = None):
    
    single_qubit_errors_ibm = [POVMs_errors_ibm['worst_case']['classical'][(qubit,)] for qubit in range(127) ]

    single_qubit_errors_rig = [POVMs_errors_rig['worst_case']['classical'][(qubit,)] for qubit in range(79) ]



    
    color_1 = color_map(0.45)#0.45
    color_2 = color_map(0.99)
    color_3 = color_map(0.1) 
    
    
    w_max = 1

    w_min = 0.00

        
    w = (w_max)/1000

    bins_new=np.arange(w_min, w_max + w, w) 
    bins_new=np.append(bins_new,np.inf)
    plt.hist(single_qubit_errors_ibm,bins=bins_new,cumulative=True,density=True,histtype='step',alpha=0.5,color = color_1,linewidth=1.5)
    plt.hist(single_qubit_errors_rig,bins=bins_new,cumulative=True,density=True,histtype='step',alpha=0.5,color = color_2,linewidth=1.5)

    labels= ["IBM Cusco","Rigetti Aspen-M-3"]
    plt.legend(labels)
    plt.title("One-qubit POVMs distance to projective meaurements  ")
    plt.xlim((-0.01,1.01))
    plt.ylim((0.0,1.1))
    if path_to_save:
        plt.savefig(path_to_save)
    else:
        plt.savefig()
    plt.clf()


def create_correlations_distance_histogram(correlations_coefficients_matrix_ibm, correlations_coefficients_matrix_rig, path_to_save=None):



    color_1 = color_map(0.45)#0.45
    color_2 = color_map(0.99)
    color_3 = color_map(0.1) 
    
    w_max = max(0, max(correlations_coefficients_matrix_rig.flatten()))

    w_min = 0.00

        
    w = (w_max)/15000

    bins_new=np.arange(w_min, w_max + w, w) 
    bins_new=np.append(bins_new,np.inf)



    plt.hist(correlations_coefficients_matrix_ibm.flatten(),bins=bins_new,cumulative=True,density=True,histtype='step',alpha=0.5,color = color_1,linewidth=1.5)
    
    plt.hist(correlations_coefficients_matrix_rig.flatten(),bins=bins_new,cumulative=True,density=True,histtype='step',alpha=0.5,color = color_2,linewidth=1.5)

    labels= ["IBM Cusco","Rigetti Aspen-M-3"]
    plt.legend(labels)
    plt.title("Cumulative histogram of correlation coefficients")
    plt.xlim((-0.01,0.41))
    plt.ylim((0.95,1.005))
   
    if path_to_save:
        plt.savefig(path_to_save)
    else:
        plt.savefig()
    plt.clf()

def create_coherence_bound_histogram(coherence_bound_dictionary_ibm:Dict, coherence_bound_dictionary_rigetti:Dict, path_to_save=None):

    ibm_data =  create_coherence_bound_histogram_data( coherence_bound_dictionary_ibm)

    rigetti_data = create_coherence_bound_histogram_data(coherence_bound_dictionary_rigetti)

    
    color_rigetti = color_map(0.45)#0.45
    color_ibm = color_map(0.99)
    w = 0.001
    bins_new=np.arange(0.03, 0.1+ w, w) 
    plt.xlim((0.03,0.1))
    plt.ylim((0,42))

    entries_ibm, edges_ibm, _= plt.hist(ibm_data,density=True,color=color_ibm,alpha=0.5,bins=bins_new)
    entries_rig, edges_rig, _= plt.hist(rigetti_data,density=False,color=color_rigetti,alpha=0.5,bins=bins_new)
    labels= ["IBM Cusco","Rigetti Aspen-M-3"]

    plt.legend(labels)
    plt.title("Histogram of bound on Coherence Strength" )
    if path_to_save:
        plt.savefig(path_to_save)
    else:
        plt.savefig()
    plt.clf()

def create_error_mitigation_prediction_histogram(noise_models_mitigated_energy_dictionary:Dict,noise_models_predicted_energy_dictionary:Dict,energy_dictionary:Dict,number_of_qubits:int, plot_title ='', path_to_save =None):
    
    noise_models_mitigated_energy_dictionary_error= compute_mitigation_errors(noise_models_mitigated_energy_dictionary['corrected_energy'],hamiltonian_energy_dictionary=energy_dictionary,number_of_qubits=number_of_qubits)
    
    benchmark_results_mean_median_dictionary = compute_mitigation_error_median_mean(mitigation_errors_dictionary=noise_models_mitigated_energy_dictionary_error,print_results=False)


    #cluster assignment with minimal median is chosen
    minimal_median_cluster_assignment=min(benchmark_results_mean_median_dictionary['median'],key = benchmark_results_mean_median_dictionary['median'].get)

    #product noise model data are red from mitigation results
    product_model_error_list = list(noise_models_mitigated_energy_dictionary_error[next(iter(noise_models_mitigated_energy_dictionary_error.keys()))].values())

    #best (i.e. with minimal median) noise model data are red from mitigation results 
    best_cn_model_error_list = list(noise_models_mitigated_energy_dictionary_error[minimal_median_cluster_assignment].values())

    #plot colors are set
    color_product_model = color_map(0.45)#0.45
    color_cn_best_model = color_map(0.99)
    raw_error_color = color_map(0.1) 

    #raw error computation
    raw_errors_list = []
    error_prediction_best_cn =[]
    error_tpm = []  
    tpm_clusters = tuple(next(iter(noise_models_predicted_energy_dictionary["predicted_energy"].keys() ))) 
    for key, hamiltonian_energy_dictionary in energy_dictionary.items():
        raw_errors_list.append(abs(hamiltonian_energy_dictionary['energy_raw']-hamiltonian_energy_dictionary['energy_ideal'])/number_of_qubits)
        error_prediction_best_cn.append(abs(hamiltonian_energy_dictionary['energy_raw']- noise_models_predicted_energy_dictionary["predicted_energy"][minimal_median_cluster_assignment][key]["predicted_energy"])/number_of_qubits)
        error_tpm.append(abs(hamiltonian_energy_dictionary['energy_raw']- noise_models_predicted_energy_dictionary["predicted_energy"][tpm_clusters][key]["predicted_energy"])/number_of_qubits)

    
    #histogram is drawn 
    color_product_model = color_map(0.45)#0.45
    color_cn_best_model = color_map(0.99)
    raw_error_color = color_map(0.1) 
    w = (max(raw_errors_list) -min(best_cn_model_error_list))/60
    bins_new=np.arange(min(best_cn_model_error_list), max(raw_errors_list) + w, w) 
    plt.hist(product_model_error_list,bins=bins_new,alpha=0.5,color = color_product_model)
    plt.hist(best_cn_model_error_list ,bins=bins_new,alpha=0.5, color = color_cn_best_model)
    plt.hist(raw_errors_list , bins=bins_new,alpha=0.5, color = raw_error_color )
    plt.ylim(0, 80)
    labels= ["TPM","CN model locality " +str(len(max(minimal_median_cluster_assignment, key=len))), "$\Delta E_{EST}(H)$"]
    plt.legend(labels)
    plt.title("Histogram of $\Delta E_{MIT}(H)$ " + plot_title)

  

    if path_to_save:
        plt.savefig(path_to_save+"error_mitigation_histogram")
    else:
        plt.savefig()
    plt.clf()

       #histogram is drawn 
    color_product_model = color_map(0.45)#0.45
    color_cn_best_model = color_map(0.99)
    raw_error_color = color_map(0.1) 
    w = (max(raw_errors_list) -min(best_cn_model_error_list))/60
    bins_new=np.arange(min(best_cn_model_error_list), max(raw_errors_list) + w, w) 
    plt.hist(error_tpm,bins=bins_new,alpha=0.5,color = color_product_model)
    plt.hist(error_prediction_best_cn ,bins=bins_new,alpha=0.5, color = color_cn_best_model)
    plt.ylim(0, 80)
    labels= ["TPM","CN model locality " +str(len(max(minimal_median_cluster_assignment, key=len))), "$\Delta E_{EST}(H)$"]
    plt.legend(labels)
    plt.title("Histogram of $\Delta E_{PRED}(H)$ " + plot_title)

  

    if path_to_save:
        plt.savefig(path_to_save+"error_prediction_histogram")
    else:
        plt.savefig()
    plt.clf()





"""
@authors: Filip Maciejewski, Oskar Słowik, Tomek Rybotycki
@contact: filip.b.maciejewski@gmail.com
"""

from qrem.functions_qrem.functions_noise_model_heuristic import *
from qrem.noise_model_generation.CN.clustering_algorithms.base_classes_clustering.heuristic_clustering_algorithm_base import HeuristicClusteringAlgorithmBase
import numpy as np


class AlphaScanError(Exception):
    """Class for handling situation where no primising plateaus are detected during alpha scan"""
    pass


def get_avg_correlation(correlations_table):
    sum = 0
    qubits = len(correlations_table)
    no_of_corr = qubits * qubits
    for row in correlations_table:
        for corr in row:
            sum += corr
    return sum / no_of_corr


# C_max has to be at least 2
def get_avg_large_corr(correlations_table, C_maxsize):
    qubits = correlations_table.shape[0]
    no_of_large_corrs = qubits * (C_maxsize - 1)
    flat = correlations_table.flatten()
    flat[::-1].sort(kind='heapsort')
    # print(flat)
    largest_corrs = flat[:no_of_large_corrs]

    sum = 0
    for corr in largest_corrs:
        sum += corr
    return sum / no_of_large_corrs


def compare_clusterings(clustering_1, clustering_2):
    isEqual = False

    los1 = set([frozenset(x) for x in clustering_1])
    los2 = set([frozenset(x) for x in clustering_2])

    # print(los1)
    # print(los2)

    if los1 == los2:
        isEqual = True
    return isEqual


'''
cl1=[[1,2],[3,4,5]]
cl2=[[4,3,5],[2,1]]
print(compare_clusterings(cl1,cl2))
'''


def plateaux_detection(hca: HeuristicClusteringAlgorithmBase):
    avg_large_corr = get_avg_large_corr(hca.correlations_table, hca.C_maxsize)
    # print(avg_large_corr)

    first = True
    c_old = []
    incr = 0.01
    stability_intervals = []
    on_plateaux = False
    for coeff in tqdm(np.arange(0, 2 + incr, incr)):
        alpha = coeff * avg_large_corr
        c, cost = hca.clusterize(alpha=alpha)
        if (not first):
            if (compare_clusterings(c, c_old)):
                if (not on_plateaux):
                    on_plateaux = True
                    plateaux_start = coeff - incr
                # print("Stability on v detected at: "+str(coeff))
                # print(c)
            else:
                if (on_plateaux):
                    on_plateaux = False
                    plateaux_end = coeff - incr
                    stability_intervals.append((plateaux_start, plateaux_end))

        c_old = c
        first = False

    return avg_large_corr, stability_intervals


def suggest_alpha_via_scan(hca: HeuristicClusteringAlgorithmBase, values_to_return=None):
    avg_large_corr, stability_intervals = plateaux_detection(hca)
    print("avg_large_corr: "+str(avg_large_corr)+"stability_intervals: "+str(stability_intervals))
    if (len(stability_intervals) < 2):
        raise AlphaScanError("No promising plateas detected")

    promising_plateaux = tuple([avg_large_corr*x for x in stability_intervals[1]])
    if values_to_return is None:
        return promising_plateaux
    else:
        alpha_min = promising_plateaux[0]
        alpha_max = promising_plateaux[1]
        interval = (alpha_max - alpha_min) / (values_to_return - 1)
        return np.arange(alpha_min, alpha_max + interval, interval)



"""
Created on 23.08.2021

@author: Oskar Słowik
@contact: osslowik@gmail.com
"""
from qrem.noise_model_generation.CN.clustering_algorithms.clustering_methods.child_classes.heuristic_clustering_algorithm_base import HeuristicClusteringAlgorithmBase
from qrem.noise_model_generation.CN.clustering_algorithms.clustering_methods.functions.functions_noise_model_heuristic_help import get_avg_large_corr, suggest_alpha_via_scan


#JT class used when performing benchmarks to set clustering algorithm, probably can be joined with HeuristicClusteringAlgorithmBase
class HeuristicClusteringAlgorithm(HeuristicClusteringAlgorithmBase):

    def __init__(self, **kwargs):
        super(HeuristicClusteringAlgorithm, self).__init__(**kwargs)

    def suggest_alpha_heuristic(self):
        if self.correlations_table is None:
            raise ValueError("correlations_table not specified.")
        if self.C_maxsize is None:
            raise ValueError("C_maxsize not specified.")
        self.alpha = 0.5 * get_avg_large_corr(self.correlations_table, self.C_maxsize)

    def suggest_alpha_scan(self, values_to_return=None):
        self.alpha = suggest_alpha_via_scan(self, values_to_return)



"""
@authors: Filip Maciejewski, Oskar Słowik, Tomek Rybotycki
@contact: filip.b.maciejewski@gmail.com

REFERENCES:
[0] Filip B. Maciejewski, Zoltán Zimborás, Michał Oszmaniec,
"Mitigation of readout noise in near-term quantum devices
by classical post-processing based on detector tomography",
Quantum 4, 257 (2020)

[0.5] Filip B. Maciejewski, Flavio Baccari, Zoltán Zimborás, Michał Oszmaniec,
"Modeling and mitigation of cross-talk effects in readout noise 
with applications to the Quantum Approximate Optimization Algorithm",
Quantum 5, 464 (2021).

"""

from typing import Optional, List, Dict, Union

import numpy as np
from qrem.functions_qrem import ancillary_functions as anf
from qrem.noise_characterization.tomography_design.overlapping import overlapping_tomography_functions as otf

class LabelsBaseDDOT:
    def __init__(self,
                 number_of_qubits: int,
                 subsets: Optional[Union[Dict[str, List[int]], List[List[int]]]] = None,
                 maximal_circuits_amount: Optional[int] = 1500,
                 show_progress_bars: Optional[bool] = True,
                 ):
        self._number_of_qubits = number_of_qubits
        self._subsets = subsets
        self._maximal_circuits_amount = maximal_circuits_amount
        self._circuits_list = []

        self._show_progress_bars = show_progress_bars



    @property
    def subsets(self) -> Union[Dict[str, List[int]], List[List[int]]]:
        return self._subsets

    @subsets.setter
    def subsets(self, subsets: Union[Dict[str, List[int]], List[List[int]]]) -> None:
        self._subsets = subsets

    @property
    def circuits_list(self) -> List[List[int]]:
        return self._circuits_list

    @circuits_list.setter
    def circuits_list(self, circuits_list: List[List[int]]) -> None:
        self._circuits_list = circuits_list
    # @staticmethod
    # def get_qubits_keystring(qubits_list):
    #     #TODO FBM: outdated, change it
    #     anf.get_qubits_keystring(list_of_qubits=qubits_list)

    def get_random_circuit(self,
                           number_of_qubits:int=None)->List[int]:

        if number_of_qubits is None:
            number_of_qubits = self._number_of_qubits
        return [np.random.randint(0, 2) for _ in
                range(number_of_qubits)]

    def get_random_circuit_with_fixed_state_of_some_qubits(self,
                                                           fixed_states: Dict[int, int])->List[int]:
        """
        :param fixed_states: dictionary where each KEY is index of qubit,
                            and VALUE denotes qubit's state (0 or 1)
        :return:
        """

        fixed_qubits = fixed_states.keys()

        circuit = []
        for qubit_index in range(self._number_of_qubits):
            if qubit_index in fixed_qubits:
                circuit.append(fixed_states[qubit_index])
            else:
                circuit.append(np.random.randint(0, 2))

        return circuit

    def add_circuits(self,
                     circuits: List[List[str]]) -> None:
        for circuit in circuits:
            self._circuits_list.append(circuit)

    def get_random_circuits_list(self,
                                 number_of_circuits: int,
                                 fixed_states: Optional[Dict[int, int]] = None)->List[List[int]]:
        if fixed_states is None:
            return [self.get_random_circuit() for _ in range(number_of_circuits)]
        else:
            return [self.get_random_circuit_with_fixed_state_of_some_qubits(fixed_states) for _ in
                    range(number_of_circuits)]

    def add_random_circuits(self,
                            number_of_circuits: int,
                            fixed_states: Optional[Dict[int, int]] = None):

        self.add_circuits(self.get_random_circuits_list(number_of_circuits, fixed_states))

    def get_random_collection(self,
                              number_of_circuits: int,
                              locality: int,
                              desired_accuracy: float = 0.005):
        # TODO FBM finish

        1

        # suggested_number_of_circuits.

    def get_parallel_tomography_on_subsets(self,
                                           number_of_circuits: int,
                                           non_overlapping_subsets:Optional[List[List[int]]]=None,
                                           ) -> List[List[int]]:
        """

        :param number_of_circuits: should be power of 2
        :param subsets:
        :param number_of_symbols:
        :return:
        :rtype:
        """

        if non_overlapping_subsets is None:
            non_overlapping_subsets = self._subsets

        number_of_symbols = 2

        return otf.get_parallel_tomography_on_non_overlapping_subsets(number_of_circuits=number_of_circuits,
                                                                      number_of_symbols=number_of_symbols,
                                                                      non_overlapping_subsets=non_overlapping_subsets)


"""
@author: Jan Tuziemski, Filip Maciejewski, Joanna Majsak, Oskar Słowik, Marcin Kotowski, Katarzyna Kowalczyk-Murynka, Paweł Przewłocki, Piotr Podziemski, Michał Oszmaniec
@contact: filip.b.maciejewski@gmail.com
"""

from collections import defaultdict
from typing import Dict, Tuple, List, Union, Optional

from tqdm import tqdm
import numpy as np
from qrem.common import convert, utils

from qrem.functions_qrem import quantum_ancillary_functions as quanf, ancillary_functions as anf
from qrem.functions_qrem import ground_state_approximations as gsa


#(PP) dobuled in other files
#(PP) part of a model of Bravi noise
#(PP) should go to qrem.noise_modelling
def _return_1local_gen(operator_type):
    if operator_type == '0':
        gen = [[-1, 0],
               [1, 0]]
    elif operator_type == '1':
        gen = [[0, 1],
               [0, -1]]
    else:
        raise ValueError('Wrong generator type!')
    return np.array(gen)


#(PP) dobuled in other files
#(PP) part of a model of Bravi noise
#(PP) should go to qrem.noise_modelling
def _return_2local_gen(operator_type):
    if operator_type == '00':
        gen = [[-1, 0, 0, 0],
               [0, 0, 0, 0],
               [0, 0, 0, 0],
               [1, 0, 0, 0]]
    elif operator_type == '01':
        gen = [[0, 0, 0, 0],
               [0, -1, 0, 0],
               [0, 1, 0, 0],
               [0, 0, 0, 0]]
    elif operator_type == '10':
        gen = [[0, 0, 0, 0],
               [0, 0, 1, 0],
               [0, 0, -1, 0],
               [0, 0, 0, 0]]
    elif operator_type == '11':
        gen = [[0, 0, 0, 1],
               [0, 0, 0, 0],
               [0, 0, 0, 0],
               [0, 0, 0, -1]]
    else:
        raise ValueError('Wrong generator type!')
    return np.array(gen)


def __convert_rates_dictionary_into_hamiltonian(ctmp_rates_dictionary,
                                                number_of_qubits):
    constant = 0
    hamiltonian_weights = defaultdict(float)
    for qubits_subset, rates_dic in ctmp_rates_dictionary.items():
        if len(qubits_subset) == 1:
            r0, r1 = rates_dic['0'], rates_dic['1']
            constant += (r0 + r1) / 2
            hamiltonian_weights[qubits_subset] += (r0 - r1) / 2


        elif len(qubits_subset) == 2:
            r00, r01, r10, r11 = rates_dic['00'], rates_dic['01'], rates_dic['10'], rates_dic['11']

            constant += (r00 + r01 + r10 + r11) / 4

            hamiltonian_weights[(qubits_subset[0],)] += (r00 + r01 - r10 - r11) / 4
            hamiltonian_weights[(qubits_subset[1],)] += (r00 - r01 + r10 - r11) / 4
            hamiltonian_weights[qubits_subset] += (r00 - r01 - r10 + r11) / 4

    return hamiltonian_weights, constant


def __get_G_energy_for_fixed_state(input_state: Union[str, List[Union[str, int]]],
                                   ctmp_rates_dictionary: Dict[Tuple[int], Dict[str, float]],
                                   number_of_qubits: int,
                                   qubits_mapping: Dict[int, int] = None):
    if qubits_mapping is None:
        qubits_mapping = {qi: qi for qi in range(number_of_qubits)}

    part_1q, part_2q = 0, 0
    for subset, local_dict in ctmp_rates_dictionary.items():

        if len(subset) == 1:
            qi = qubits_mapping[subset[0]]
            part_1q += ctmp_rates_dictionary[subset][input_state[qi]]
        elif len(subset) == 2:
            qi, qj = qubits_mapping[subset[0]], qubits_mapping[subset[1]]
            part_2q += ctmp_rates_dictionary[subset][input_state[qi] + input_state[qj]]
        else:
            raise ValueError(f"Wrong rates dictionary with key: 'subset'.")

    return part_1q + part_2q


def return_generator_matrix(operator_type):
    if len(operator_type) == 1:
        return _return_1local_gen(operator_type=operator_type)
    elif len(operator_type) == 2:
        return _return_2local_gen(operator_type=operator_type)
    else:
        raise ValueError('Wrong generator dimension in r_dict!')


def get_generators_dictionary_from_rates(rates_dictionary):
    generators_dictionary = {}
    for qubits_tuple in rates_dictionary.keys():
        operator_now = 0
        for operator_type in rates_dictionary[qubits_tuple].keys():
            operator_now += return_generator_matrix(operator_type=operator_type) * \
                            rates_dictionary[qubits_tuple][operator_type]

        generators_dictionary[qubits_tuple] = operator_now

    return generators_dictionary


def get_global_generator_from_rates(rates_dictionary: Dict[Tuple[int], Dict[str, float]],
                                    number_of_qubits: int):
    global_noise_matrix = np.zeros((2 ** number_of_qubits, 2 ** number_of_qubits), dtype=float)

    for qubits_tuple in rates_dictionary.keys():
        subset_size = len(qubits_tuple)
        local_dimension = int(2 ** subset_size)

        local_generator_now = np.zeros((local_dimension, local_dimension), dtype=float)
        for operator_type in rates_dictionary[qubits_tuple]:
            error_rate = rates_dictionary[qubits_tuple][operator_type]

            local_generator_now += error_rate * return_generator_matrix(operator_type=operator_type)

        # raise KeyError
        generator_now_embeded = quanf.embed_operator_in_bigger_hilbert_space(
            number_of_qubits=number_of_qubits,
            local_operator=local_generator_now,
            global_indices=qubits_tuple).real

        global_noise_matrix += generator_now_embeded

    return global_noise_matrix


def construct_model_from_rates(rates_dictionary):
    generators_dictionary = {}
    for qubits_tuple in rates_dictionary.keys():
        subset_size = len(qubits_tuple)
        local_dimension = int(2 ** subset_size)

        local_generator_now = np.zeros((local_dimension, local_dimension), dtype=float)
        for operator_type in rates_dictionary[qubits_tuple]:
            error_rate = rates_dictionary[qubits_tuple][operator_type]

            local_generator_now += error_rate * return_generator_matrix(operator_type=operator_type)

        generators_dictionary[qubits_tuple] = local_generator_now
    return generators_dictionary


def _find_optimal_noise_strength_CTMP_bruteforce(
        ctmp_rates_dictionary: Dict[Tuple[int], Dict[str, float]],
        number_of_qubits: int,
        qubits_mapping: Optional[Dict[int, int]] = None):
    if qubits_mapping is None:
        qubits_mapping = {qi: qi for qi in range(number_of_qubits)}

    gamma = 10 ** 6
    gamma_max = 0

    integers_range = range(2 ** number_of_qubits)

    if number_of_qubits > 10:
        integers_range = tqdm(integers_range)

    for integer in integers_range:
        bitstring = convert.integer_to_bitstring(integer=integer, number_of_bits=number_of_qubits)

        potential_gamma = __get_G_energy_for_fixed_state(input_state=bitstring,
                                                         ctmp_rates_dictionary=ctmp_rates_dictionary,
                                                         number_of_qubits=number_of_qubits,
                                                         qubits_mapping=qubits_mapping
                                                         )

        if potential_gamma < gamma:
            gamma = potential_gamma

        if potential_gamma > gamma_max:
            gamma_max = potential_gamma
    # print(gamma, gamma_max)
    return gamma, gamma_max


def _find_optimal_noise_strength_CTMP_SDP_relaxation(
        ctmp_rates_dictionary: Dict[Tuple[int], Dict[str, float]],
        number_of_qubits: int,
        hierarchy_level: Optional[int] = 1) -> Tuple[float]:
    hamiltonian_dictionary, constant = __convert_rates_dictionary_into_hamiltonian(
        ctmp_rates_dictionary=ctmp_rates_dictionary,
        number_of_qubits=number_of_qubits)

    lower_bound, upper_bound = gsa.find_ground_state_with_SDP_relaxations(
        weights_dictionary=hamiltonian_dictionary,
        number_of_qubits=number_of_qubits,
        hierarchy_level=hierarchy_level)

    return lower_bound + constant, upper_bound + constant


def get_CTMP_error_rates_sum(rates_dictionary: Dict[Tuple[int], Dict[str, float]]):
    noise_strength_CTMP = 0
    for subset in rates_dictionary.keys():
        for operator_type in rates_dictionary[subset].keys():
            noise_strength_CTMP += rates_dictionary[subset][operator_type]

    return noise_strength_CTMP


def find_optimal_noise_strength_CTMP(ctmp_rates_dictionary: Dict[Tuple[int], Dict[str, float]],
                                     number_of_qubits: int,
                                     method_name='SDP_relaxation',
                                     method_kwargs=None):
    if method_name.upper() == 'BRUTEFORCE':
        if method_kwargs is None:
            method_kwargs = {'qubits_mapping': None}
        _, noise_strength_CTMP = _find_optimal_noise_strength_CTMP_bruteforce(
            ctmp_rates_dictionary=ctmp_rates_dictionary,
            number_of_qubits=number_of_qubits,
            **method_kwargs)
    elif method_name.upper() == 'SDP_RELAXATION':
        if method_kwargs is None:
            method_kwargs = {'hierarchy_level': 1}

        _, noise_strength_CTMP = _find_optimal_noise_strength_CTMP_SDP_relaxation(
            ctmp_rates_dictionary=ctmp_rates_dictionary,
            number_of_qubits=number_of_qubits,
            **method_kwargs)

    elif method_name.upper() == 'SUM_OF_RATES':
        noise_strength_CTMP = get_CTMP_error_rates_sum(rates_dictionary=ctmp_rates_dictionary)
    else:
        raise ValueError(f"Method name '{method_name}' not recognized.\n "
                         f"Please choose one of the following:"
                         f" - 'SDP_relaxation' "
                         f" - 'bruteforce' "
                         f" - 'sum_of_rates' ")
    return noise_strength_CTMP


def get_pairs_complement_dictionary_from_clusters_list(clusters_list: List[Tuple[int]]) -> Dict[
    Tuple[int], List[int]]:
    """

    :param clusters_list: List of clusters in the noise model
    :type clusters_list:
    :return:
    :rtype:
    """

    pairs_complements_dictionary = {}
    number_of_qubits = int(np.max([np.max(cl) for cl in clusters_list]))
    qubit_to_cluster_map = {}
    for cluster in clusters_list:
        for qubit_index in cluster:
            qubit_to_cluster_map[(qubit_index,)] = cluster

    all_qubits = list(range(number_of_qubits))
    for qi in range(number_of_qubits):
        cluster = qubit_to_cluster_map[(qi,)]
        not_cluster = list(set(all_qubits).difference(set(cluster)))

        for qj in not_cluster:
            pairs_complements_dictionary[tuple(sorted((qi, qj)))] = []

        for qj in cluster:
            if qj > qi:
                pairs_complements_dictionary[(qi, qj)] = list(set(cluster).difference(set((qi, qj))))
    return pairs_complements_dictionary


def get_theoretical_overhead_CTMP(noise_strength_CTMP: float, ):
    return np.exp(4 * noise_strength_CTMP)


def get_theoretical_number_of_samples_CTMP(noise_strength_CTMP: float,
                                           target_estimation_accuracy: float):
    return np.ceil(4 * np.exp(4 * noise_strength_CTMP) / target_estimation_accuracy ** 2)


def _convert_to_local_noise_rates_dictionary(CTMP_rates_dictionary,
                                             clusters_list: List[Tuple[int]]):
    local_rates_dictionaries = {cluster: {} for cluster in clusters_list}
    for cluster in clusters_list:
        pairs_in_cluster = [(qi, qj) for qi in cluster for qj in cluster if qj > qi]
        cluster_mapping = utils.map_index_to_order(cluster)

        for (qi, qj) in pairs_in_cluster:
            local_rates_dictionaries[cluster][(cluster_mapping[qi],
                                               cluster_mapping[qj])] = CTMP_rates_dictionary[
                (qi, qj)]

        for q0 in cluster:
            local_rates_dictionaries[cluster][(cluster_mapping[q0],)] = \
                CTMP_rates_dictionary[(q0,)]

    return local_rates_dictionaries


def calculate_noise_strengths_local(
        local_rates_dictionaries,
        method='bruteforce'):


    local_noise_strengths = {}
    for cluster, local_rates_dictionary in local_rates_dictionaries.items():
        noise_strength_CTMP = find_optimal_noise_strength_CTMP(
            ctmp_rates_dictionary=local_rates_dictionary,
            number_of_qubits=len(cluster),
            method_name=method)

        local_noise_strengths[cluster] = noise_strength_CTMP

    return local_noise_strengths


def get_local_noise_matrices(
        local_rates_dictionaries,
        local_noise_strengths
):
    local_generators_dictionary_nested = {}

    clusters_list = list(local_rates_dictionaries.keys())

    for cluster, local_rates_dictionary in local_rates_dictionaries.items():
        local_generators_dictionary_nested[cluster] = get_generators_dictionary_from_rates(
            rates_dictionary=local_rates_dictionary)

    local_noise_matrices = {}
    for cluster in clusters_list:
        # enumerated_indices = utils.map_index_to_order(cluster)
        number_of_qubits_local = len(cluster)
        iden = np.eye(2 ** len(cluster), dtype=float)

        local_generator = 0
        local_generators_now = local_generators_dictionary_nested[cluster]
        for key, value in local_generators_now.items():
            local_generator += quanf.embed_operator_in_bigger_hilbert_space(
                number_of_qubits=number_of_qubits_local,
                local_operator=value,
                global_indices=key).real

        local_noise_matrices[cluster] = abs(
            iden + 1 / local_noise_strengths[cluster] * local_generator)

    return local_noise_matrices



import numpy as np
import random
from typing import Dict, Tuple, List
from qrem.types import CTMPModelData
from qrem.ctmp.simtools.sampling_utils import check_transition, apply_transition



"""
Converts a description of an observable obs given by coefficients in Z's to a numpy array; 
    last column = coefficient.

E.g. [ [(), 1], [(0,1,3), -2]] denotes O = I - 2 * ZZIZ and is converted into:
0, 0, 0, 0, 1
1, 1, 0, 1, -2

Z - list of coefficients in the above format
n - number of qubits
"""
def convert(obs: List[List], n: int) -> np.array:
    result = np.zeros((len(obs), n+1))
    for k in range(len(obs)):
        result[k, obs[k][0]] = 1
        result[k, -1] = obs[k][1]
    return result

"""
Computes obs(x), where obs is a numpy array describing an observable as above and x is an array of 0 and 1
"""
def evaluate(obs: List[List], x: np.array) -> float:
    parities = np.einsum("ij,j->i", obs[:,:-1], x).astype(int) % 2
    return ((-1)**parities * obs[:,-1]).sum()


"""
Return a sample x (array of 0s and 1s) from the distribution <x| B^alpha |s>, where B = I + gamma^{-1}G
input_state - |s>, string of 0s and 1s
"""
def sample(model: CTMPModelData, alpha: int, input_state: str) -> str:
    gamma = model.gamma
    current_state = input_state
    for t in range(alpha):
        transitions = list(filter(lambda x: check_transition(current_state, x), model.rates))
        if len(transitions) > 0:
            transition_rates = [error[4] for error in transitions]
            total_rate = sum(transition_rates)
            p = 1 - total_rate / gamma # probability of staying in place according to B
            if p < 0:
                raise Exception("Invalid gamma parameter - transition matrix not stochastic")
            if np.random.uniform() > p:
                # if not staying in place, pick random transition
                k = random.choices(range(len(transition_rates)), [x/total_rate for x in transition_rates])[0]
                current_state = apply_transition(current_state, transitions[k])
    return current_state

"""Arguments:
x - bit string of size n
indices - tuple representing a set of indices S

Returns a 2^|S| vector with 1 at index corresponding to position y such that x restricted to S equals y
and 0 otherwise.
"""
def delta(x: str, indices: Tuple[int]) -> np.array:
    result = np.zeros(2**len(indices))
    pos = int(''.join([x[i] for i in indices]), 2)
    result[pos] = 1
    return result

"""Returns a random element from probability distribution given by counts, where e.g.
counts = {"000": 70, "001": 123, ...}
"""
def random_element(counts: Dict[str, int]) -> str:
    s = sum(counts.values())
    d = {k: v/float(s) for k, v in counts.items()}
    return np.random.choice(list(d.keys()), 1, p=list(d.values()))[0]




import numpy as np

from typing import Tuple, Dict, List, Optional






#This fuction takes a dictionary consisting of DOT experiment setting (string describing unitaries) and result and transforms it accordingly to a given noise model
#The noise model is specified accordingly to QREM convection, ditionary where keay are tuples of integers encoding qubits belonging to a cluster and assosiated array of noise matrices on clusters (key avaraged indicates absence of neighborhoods )




import random as rd



"""
A function creating a string of outcomes from measurement statistics

input parameters:

marginal_results_list - a list with marginal measurement outcomes results are encoded in intgeres, eg [0,3,0,1,1,2] is a list of 6 results for a two qubit marginals 

"""

def transform_marginal_statistics_to_list_of_results(marginal_results):
    
    results_number= len(marginal_results)
    
    marginal_results_list=[]
    
    for i in range(results_number):
        for j in range(marginal_results[i]):
            marginal_results_list.append(i) 
    rd.shuffle(marginal_results_list)

    return marginal_results_list





"""
A function that creates a joint result list by gluing outcomes from two marginals lists 

input parameters:  

marginal_list_1 - a list contaning marinal results, results are encoded in intgeres, eg [0,3,0,1,1,2] is a list of 6 results for a two qubit marginals 

marginal_list_2 - as above 

output parameters:

results_list - a join list of results



"""

'''
A creating a joint sample out of samples created according to product probability distributions

input parameters:  


marginal_results_list - a list of list corresponding to product samples


output parameters:

joint_sample - a list corresponding to the joint sample


'''

def merge_samples_list(marginal_list_1, dim1, marginal_list_2, dim2):
    


    results_list = [0 for i in range(dim1*dim2)]

    for r1, r2 in zip(marginal_list_1,marginal_list_2):
        results_list[r1*dim2+r2]+=1

    return results_list









#
def create_joint_measurement_statisctic_from_marginals(marginal_results_list):
    
    joint_samplec=[1]

    for el in marginal_results_list:
        joint_sample = np.kron(joint_sample,el)
    
    return joint_sample 


'''
A function simulating adding noise to noiseless results of simulations

input parameters:  


results_dictionary - dictionary with results of a perfect DOT experiment

noise_model_description - a dictionary:
        keys - tuples corresponding to cluster assigments
        values - a dictionary with keys - 'averaged' - averaged noise matrices for clusters 
        
        neighborhoods_tuples - this is not used now, but can be used in the future when neighbors are added   

'''


def simulate_noise_results_dictionary(results_dictionary: Dict[str, Dict[str, int]],
                            noise_model_description: Dict[Tuple[int], Dict[str, np.ndarray]],
                            neighborhoods_tuples: Optional[Dict[Tuple[int], List[int]]]={}
                            ):

    #_noisy_results_total - a dictionary that holds results of noisy experiments, at the beginning empty

    noisy_results_total={}
    # an outer loop over settings and results from results_dictionary, a dictionary storing results of ideal (noisless) experiment
    for settings, results_for_settings in results_dictionary.items():

        #an array for new noise samples, noise is added sample by sample, the qubit ordering according to the qubit cluster list
        
        new_results=np.array([ 0 for _ in range(2**len(list(results_for_settings.keys())[0]))])

        # for each input state a probability distribution encoding noise is created

        for input_state, number_of_samples in results_for_settings.items():

            

            # this loop goes over clusters and respective noise matrices, as for now we support only the simplified model without nieghbours

            prob_dist_now =[]
            marginal_samples =[]
            first=True
            dim_total=1
            for cluster_qubits, noise_matrices_on_cluster in noise_model_description.items():
                
                

                dim_now=2**len(cluster_qubits)

               

            
                
                #a state of qubits belonging to the current cluster is established

                input_on_cluster = ''.join([input_state[x] for x in cluster_qubits])

                #the code below is prepared for clusters extension, which is not implemented now
                #TODO: FINISH this
                
                #cluster noise matrix is specified
                noise_matrix = noise_matrices_on_cluster['averaged'] 
                        

                # a transfer probability is chosen for a given state of the results

                prob_dist_now=noise_matrix[:, int(input_on_cluster, 2)]

                
         

                #samples are created according to respective product probabiity distributions
                # the if statment is needed to handle the case when a first sample is crated (there are no other samples to join)
                # the general idea of joining different samples is the following
                # 1) a list of counts is crated according to a probability distribution e.g. for two qubits it is of a formm [4,8,1,7] 
                # for an experiment with 20 shots
                # 2) The list is tranformed to a randomized result list e.g [2,0,1,3,...]
                # 3) This list is joined with a previously generated measurement statistic
                                  
                if first:
                    marginal_samples = np.random.multinomial(n=number_of_samples,pvals=prob_dist_now)
                    marginal_samples=transform_marginal_statistics_to_list_of_results(marginal_samples)
                    
                    dim_total=dim_total*dim_now
                    
                    first=False
                else:

                      

                    marginal_samples_temp=transform_marginal_statistics_to_list_of_results(np.random.multinomial(n=number_of_samples,pvals=prob_dist_now))

                   
                    marginal_results=merge_samples_list(marginal_samples,dim_total,marginal_samples_temp,dim_now)

                    marginal_samples = transform_marginal_statistics_to_list_of_results(marginal_results)

                    dim_total=dim_total*dim_now

            #When the loop over clusters is finished new noisy result is updated 

            
            new_results =new_results + marginal_results
   




           



            

        #In the above, the orders of qubits was that of clusters. Here the initial order of qubits is restored.

        noisy_results={}

        # this is a loop over all possible bitstrings (results) of qubits
        #the order of qubits is changed
        #key on clusters is a string corresponding to results

        key = ['0' for _ in  range(len(input_state))]

        for i in range(len(new_results)):

            #the bit string corresponding to results is created

            key_on_clusters = "{0:b}".format(2**len(input_state)+i)
            key_on_clusters = key_on_clusters[1:]

            #counter is a variable that goes over physical qubits

            counter=0

            #the first for loop is over different cluster assigments

            for qubits_in_clusters in noise_model_description.keys():

                #the second for loop is over qubits belonging to that cluster

                for qubit in qubits_in_clusters:
                    #order of qubits is corrected: in the noisy results qubit on counter place, in put on the place according to the qubit ordering
                    key[qubit] = key_on_clusters[counter]
                    counter+=1

            #now it is checked wheter a given result appers in the noise results, if yess, it is assigned to a dictionary

            if new_results[i] != 0:

                noisy_results.update({''.join(key):new_results[i]})

        noisy_results_total.update({settings:noisy_results})

    return noisy_results_total






import operator
from functools import reduce
from qrem.types import CNModelData
from typing import Dict, Tuple, List
import numpy as np
from qrem.common import probability
from qrem.common import math
from qrem.functions_qrem import povmtools 
from qrem.functions_qrem import functions_data_analysis as fda





#TODO mitigate_single_marginal

def mitigate_marginal(marginal: Tuple, results_dictionary: Dict[str, Dict[str, int]], noise_model:type[CNModelData],threshold:float,ensure_proper_probability_distribution:bool = False, check_inverse_noise_matrix_norm:bool = True ) ->np.array:
    
    """
    A basic function performing CN noise model based mitigation of a single marginal probability distribution
    by acting on it with an appropriate inverse noise matrix. Used in mitigation routines.

    Parameters
    ----------
    marginal : tuple
        A tuple of qubits indices specifying the marginal e.g. (0,1)   
    
    result_dictionary:
        A nested dictionary with experimental results to be mitigated. Key corresponds to bitstring encoding input state, values to a dictionary with results bitstrings and counts  
        
    noise_model : object of CNModelData class
        an object of CNModelData class storing CN noise model data

    ensure_proper_probability_distribution: bool
        a boolean specifying whether mitigated pseudo probability distribution should be projected onto proper probability distribution, by default set to False 
    
    check_inverse_noise_matrix_norm: bool
        a boolean specifying whether norm of inverse noise matrices should be checked during mitigation (the norm is connected with bound on error mitigation), by default set to True

    threshold: float
        sets threshold for acceptance of inverse noise matrix norm (see above), by default set to 3   


    Returns
    -------
    mitigated_marginal_probability_distribution
        a numpy array with marginal (pseudo) marginal probability distribution 


"""
    
    
    #JT: a list of clusters that are involved in marginal is created
    clusters_in_marginal_list = noise_model.get_clusters_in_marginal_list(marginal=marginal)
    
    #old
    #get_clusters_in_marginal_list(marginal=marginal, noise_model=noise_model)
    
    #creates a list of qubit indices involved in a marginal
    unordered_qubits_in_marginal_list = list(reduce(operator.concat, clusters_in_marginal_list))

    #qubits list is sorted

    unordered_qubits_in_marginal_list.sort()
    
    #creates an inverse noise matrix for the marginal 
    if tuple(unordered_qubits_in_marginal_list) in noise_model.inverse_noise_matrices.keys():
        clusters_inverse_noise_matrix = noise_model.inverse_noise_matrices[tuple(unordered_qubits_in_marginal_list)]
    else:
        clusters_inverse_noise_matrix = noise_model.compute_extended_inverse_noise_matrices(clusters_in_marginal_list=clusters_in_marginal_list)



    #old
    #clusters_inverse_noise_matrix = get_marginal_inverse_noise_matrix(noise_model=noise_model, clusters_in_marginal_list=clusters_in_marginal_list)

   

    #computes extended marginal probability distribution on which is 
    clusters_marginal_counts=probability.compute_marginals(results_dictionary=results_dictionary,subsets_list=[tuple(unordered_qubits_in_marginal_list)])

    #we are interested in the first key only hence the construction nex(iter()) TODO: this does not work, start from here 
    mitigated_clusters_marginal_probability_distribution = fda.convert_subset_counts_dictionary_to_probability_distribution(clusters_marginal_counts)
    
    #reshape probability vector
    mitigated_clusters_marginal_probability_distribution = np.array(mitigated_clusters_marginal_probability_distribution).reshape(mitigated_clusters_marginal_probability_distribution.shape[0], 1)

    #performs mitigation on an extended marginal 

    #if the norm does not exceed the threshold, the mitigation is performed 
    if check_inverse_noise_matrix_norm:
        if np.linalg.norm(clusters_inverse_noise_matrix, ord=1) < threshold:

            mitigated_clusters_marginal_probability_distribution = clusters_inverse_noise_matrix.dot(mitigated_clusters_marginal_probability_distribution)

    #if there is no verification of the norm condition, mitigation is performed 
    else:

        mitigated_clusters_marginal_probability_distribution = clusters_inverse_noise_matrix.dot(mitigated_clusters_marginal_probability_distribution)

     


   
        

    #returns marginalization of an  extended marginal

    mitigated_marginal_probability_distribution = probability.compute_marginal_of_probability_distribution(mitigated_clusters_marginal_probability_distribution,[unordered_qubits_in_marginal_list.index(qubit) for qubit in marginal])

    if ensure_proper_probability_distribution:
        if not probability.is_valid_probability_vector(mitigated_marginal_probability_distribution):
           mitigated_marginal_probability_distribution = povmtools.find_closest_prob_vector_l2(mitigated_marginal_probability_distribution).flatten()

    
    return mitigated_marginal_probability_distribution


def mitigate_marginal_state_dependent(marginal: Tuple, results_dictionary: Dict[str, Dict[str, int]], noise_model:type[CNModelData],threshold:float,ensure_proper_probability_distribution:bool = False, check_inverse_noise_matrix_norm:bool = True ) ->np.array:
    
    """
    A basic function performing CN noise model based mitigation of a single marginal probability distribution
    by acting on it with an appropriate inverse noise matrix. Used in mitigation routines.

    Parameters
    ----------
    marginal : tuple
        A tuple of qubits indices specifying the marginal e.g. (0,1)   
    
    result_dictionary:
        A nested dictionary with experimental results to be mitigated. Key corresponds to bitstring encoding input state, values to a dictionary with results bitstrings and counts  
        
    noise_model : object of CNModelData class
        an object of CNModelData class storing CN noise model data

    ensure_proper_probability_distribution: bool
        a boolean specifying whether mitigated pseudo probability distribution should be projected onto proper probability distribution, by default set to False 
    
    check_inverse_noise_matrix_norm: bool
        a boolean specifying whether norm of inverse noise matrices should be checked during mitigation (the norm is connected with bound on error mitigation), by default set to True

    threshold: float
        sets threshold for acceptance of inverse noise matrix norm (see above), by default set to 3   


    Returns
    -------
    mitigated_marginal_probability_distribution
        a numpy array with marginal (pseudo) marginal probability distribution 


"""
    
    
    #JT: a list of clusters that are involved in marginal is created
    clusters_and_neighbors_in_marginal_dictionary = noise_model.get_clusters_and_neighborhoods_in_marginal_dictionary(marginal=marginal)
    
    #old
    #get_clusters_in_marginal_list(marginal=marginal, noise_model=noise_model)
    
    #creates a list of qubit indices involved in a marginal
    unordered_qubits_in_marginal_list = list(reduce(operator.concat, list(clusters_and_neighbors_in_marginal_dictionary.keys())))

    
    #neighbors_list = reduce(operator.concat(list(clusters_and_neighbors_in_marginal_dictionary.values())))

    ###
    #get indices of noise matrices needed for mitigation  
    ###
    

    ###
    #compute or retrieve noise matrix form memory
    ###    

    #qubits list is sorted

    #we establish which noise matrices should be accounted for in mitigation 
    #noise_matrices_indicies = noise_model.get_noise_matrices_indexes(marginal=marginal,state=results_dictionary.keys())

    unordered_qubits_in_marginal_list.sort()
    
    #creates an inverse noise matrix for the marginal 
    #if tuple(unordered_qubits_in_marginal_list) in noise_model.inverse_noise_matrices.keys():
    #    clusters_inverse_noise_matrix = noise_model.inverse_noise_matrices[tuple(unordered_qubits_in_marginal_list)]
    #else:
    clusters_inverse_noise_matrix = noise_model.compute_extended_inverse_noise_matrices_state_dependent(clusters_in_marginal_list=list(clusters_and_neighbors_in_marginal_dictionary.keys()),marginal=marginal,state = next(iter(results_dictionary.keys())))



    #old
    #clusters_inverse_noise_matrix = get_marginal_inverse_noise_matrix(noise_model=noise_model, clusters_in_marginal_list=clusters_in_marginal_list)

   

    #computes extended marginal probability distribution on which is 
    clusters_marginal_counts=probability.compute_marginals(results_dictionary=results_dictionary,subsets_list=[tuple(unordered_qubits_in_marginal_list)])

    #we are interested in the first key only hence the construction nex(iter()) TODO: this does not work, start from here 
    mitigated_clusters_marginal_probability_distribution = fda.convert_subset_counts_dictionary_to_probability_distribution(clusters_marginal_counts)
    
    #reshape probability vector
    mitigated_clusters_marginal_probability_distribution = np.array(mitigated_clusters_marginal_probability_distribution).reshape(mitigated_clusters_marginal_probability_distribution.shape[0], 1)

    #performs mitigation on an extended marginal 

    #if the norm does not exceed the threshold, the mitigation is performed 
    if check_inverse_noise_matrix_norm:
        if np.linalg.norm(clusters_inverse_noise_matrix, ord=1) < threshold:

            mitigated_clusters_marginal_probability_distribution = clusters_inverse_noise_matrix.dot(mitigated_clusters_marginal_probability_distribution)

    #if there is no verification of the norm condition, mitigation is performed 
    else:

        mitigated_clusters_marginal_probability_distribution = clusters_inverse_noise_matrix.dot(mitigated_clusters_marginal_probability_distribution)

     


   
        

    #returns marginalization of an  extended marginal

    mitigated_marginal_probability_distribution = probability.compute_marginal_of_probability_distribution(mitigated_clusters_marginal_probability_distribution,[unordered_qubits_in_marginal_list.index(qubit) for qubit in marginal])

    if ensure_proper_probability_distribution:
        if not probability.is_valid_probability_vector(mitigated_marginal_probability_distribution):
           mitigated_marginal_probability_distribution = povmtools.find_closest_prob_vector_l2(mitigated_marginal_probability_distribution).flatten()

    
    return mitigated_marginal_probability_distribution



#correct this to allow for a results dictionary that has more than one setting

def mitigate_marginals(marginals_list: List[Tuple], results_dictionary: Dict[str, Dict[str, int]], noise_model: type[CNModelData] ,ensure_proper_probability_distribution = True, check_inverse_noise_matrix_norm:bool = False, threshold:float = 100, state_independent_mitigation:bool = False )->Dict[str, Dict[tuple[int], np.array]]:

    """
    A  function performing CN noise model based mitigation for marginal probability distributions specified by a marginals list.
    Used in mitigation routines. Mitigation is performed by multiplying marginal probability distributions by inverse noise matrices.

    Parameters
    ----------
    marginals_list : List[tuple]
        A list of tuples encoding marginals e.g. [(0,1),(4,6)]   

    result_dictionary:
        A nested dictionary with experimental results to be mitigated. Key corresponds to bitstring encoding input state, values to a dictionary with results bitstrings and counts  
        
    noise_model : object of CNModelData class
        an object of CNModelData class

    ensure_proper_probability_distribution: bool
        a boolean specifying whether mitigated pseudo probability distribution should be projected onto proper probability distribution, by default set to False 

    check_inverse_noise_matrix_norm: bool
        a boolean specifying whether norm of inverse noise matrices should be checked during mitigation (the norm is connected with bound on error mitigation), by default set to True

    threshold: float
        sets threshold for acceptance of inverse noise matrix norm (see above), by default set to 3   

    Returns
    -------

        a nested dictionary {input bitstring : {(subset_tuple): marginal_probability_distributution}}, where input bitstring encodes input setting, subset_tuple encodes qubits in the marginal,
        and marginal_probability_distributution is a (pseudo) probability distribution 
    """

    #This dictionary holds mitigated marginals   
    
   
    #A loop over experimental results dictionary and marginals specified by marginals_list, for each marginal mitigation is performed by mitigate_marginal   
    
    total_mitigated_marginals_dictionary = {}

    for result_setting in results_dictionary.keys():
        mitigated_marginals_dictionary ={}
        for marginal in marginals_list:
            
            if state_independent_mitigation:
                mitigated_marginals_dictionary[marginal] = mitigate_marginal(marginal=marginal,results_dictionary={result_setting: results_dictionary[result_setting]},noise_model=noise_model,ensure_proper_probability_distribution=ensure_proper_probability_distribution, check_inverse_noise_matrix_norm=check_inverse_noise_matrix_norm,threshold=threshold)
            
            else:
                mitigated_marginals_dictionary[marginal] = mitigate_marginal_state_dependent(marginal=marginal,results_dictionary={result_setting: results_dictionary[result_setting]},noise_model=noise_model,ensure_proper_probability_distribution=ensure_proper_probability_distribution, check_inverse_noise_matrix_norm=check_inverse_noise_matrix_norm,threshold=threshold)


        total_mitigated_marginals_dictionary[result_setting] = mitigated_marginals_dictionary
    # results are returned as a nested dictionary with key corresponding to the experimental state, and value
    # corresponds to mitigated marginals encoded in a dictionary 
    return total_mitigated_marginals_dictionary





  
    
#MOVE TO CNModelData class   

def get_marginal_inverse_noise_matrix(noise_model : type[CNModelData], clusters_in_marginal_list:List[Tuple])-> np.array:

    """
    Function determines inverse noise matrix for a given marginal. It is used as a step in mitigation_marginal function.
    The matrix is constructed as a tensor product on inverse noise matrices involved in a marginal, when it is necessary
    the matrix is permuted to ensure proper ordering of qubits indices. E.g. for a cluster_in_marginal_list 
    [(0,2),(1,3)], the returned matrix corresponds to qubits ordered as (0,1,2,3).   


    Parameters
    ----------
    noise_model : object of CNModelData class
        An object of CNModelData class 

    clusters_in_marginal_list:
        A list of tuples with clusters involved in the marginal   
        

    Returns
    -------

        An inverse noise matrix for qubits specified in clusters_in_marginal_list, qubits are sorted in ascending order 
        (e.g. for a clusters_in_marginal_list =[(0,4),(1,8)], indices of the inverse noise matrix indices correspond to qubits in the order (0,1,4,8) )
    
    """ 

    # a list of qubits is established
    unordered_qubits_in_marginal_list = list(reduce(operator.concat, clusters_in_marginal_list))

    #the inverse noise matrix is initialized as a unit numpy array 
    marginal_inverse_noise_matrix = np.array([1])
    
    # a loop over clusters in clusters_in_marginal_list
    for cluster in clusters_in_marginal_list:
    
        # total noise matrix is updated by taking tensor product with a inverse noise matrix of the currect cluster
        marginal_inverse_noise_matrix = np.kron(marginal_inverse_noise_matrix, noise_model.inverse_noise_matrices[cluster])
  
    #final noise matrix is returned, permuted to preserve ascending order of qubit indices if necessary 
    return math.permute_composite_matrix(qubits_list=unordered_qubits_in_marginal_list,noise_matrix=marginal_inverse_noise_matrix)




#MOVE TO CNModelData class   

def get_clusters_in_marginal_list(marginal : Tuple[int], noise_model : type[CNModelData]) -> List[Tuple]:
    
    """
    Function creates a list of clusters that are involved in a marginal. Used in mitigation routines.
    For a given marginal inspects provided noise model and checks clusters membership of qubits form
    marginal.
    
    Parameters
    ----------
    marginal : tuple
        A tuple specifying marginal 

    noise_model : object of CNModelData class
        An object of CNModelData class  
        

    Returns
    -------

    clusters_in_marginal_list
        A list of tuples involved in the input marginal 
        
    
    """  

    #list storing clusters involved in the marginal, empty for start
    clusters_in_marginal_list=[]
    
    #a loop over qubits in the marginal
    for qubit in marginal:
        
        #a cluster to which the qubit belongs is determined 
        cluster = noise_model.qubit_in_cluster_membership[(qubit,)]
        
        #this cluster is appended to clusters_in_marginal_list if it is not there 
        if cluster not in clusters_in_marginal_list:
            clusters_in_marginal_list.append(noise_model.qubit_in_cluster_membership[(qubit,)])
    
    #results are returned
    return clusters_in_marginal_list



##############################################
######For tests purposes######################
#############################################

def mitigate_marginal_product(marginal: Tuple, results_dictionary: Dict[str, Dict[str, int]], noise_model:type[CNModelData],ensure_proper_probability_distribution:bool = False ) ->np.array:
    
    #JT: a list of clusters that are involved in marginal is created
    clusters_in_marginal_list = get_clusters_in_marginal_list(marginal=marginal, noise_model=noise_model)

    final_marginal = np.array([1])

    for cluster in clusters_in_marginal_list:

        clusters_inverse_noise_matrix = get_marginal_inverse_noise_matrix(noise_model=noise_model, clusters_in_marginal_list=[cluster])

        clusters_marginal_counts=probability.compute_marginals(results_dictionary=results_dictionary,subsets_list=[cluster])
        
        mitigated_clusters_marginal_probability_distribution = fda.convert_subset_counts_dictionary_to_probability_distribution(clusters_marginal_counts)

        mitigated_clusters_marginal_probability_distribution =np.array(mitigated_clusters_marginal_probability_distribution).reshape(mitigated_clusters_marginal_probability_distribution.shape[0], 1)

        mitigated_clusters_marginal_probability_distribution = clusters_inverse_noise_matrix.dot(mitigated_clusters_marginal_probability_distribution)

        final_marginal = np.kron(final_marginal,mitigated_clusters_marginal_probability_distribution)
        final_marginal = final_marginal.flatten()
        if ensure_proper_probability_distribution:
            if not probability.is_valid_probability_vector(final_marginal):
                final_marginal = povmtools.find_closest_prob_vector_l2(final_marginal)
                final_marginal = final_marginal.flatten()




    
    #creates a list of qubit indices involved in a marginal
    unordered_qubits_in_marginal_list = list(reduce(operator.concat, clusters_in_marginal_list))
    
    #creates an inverse noise matrix for the marginal 
    mitigated_clusters_marginal_probability_distribution = math.permute_composite_vector(unordered_qubits_in_marginal_list, final_marginal)

    #qubits list is sorted

    unordered_qubits_in_marginal_list.sort()

   
    mitigated_marginal_probability_distribution = probability.compute_marginal_of_probability_distribution(mitigated_clusters_marginal_probability_distribution ,[unordered_qubits_in_marginal_list.index(qubit) for qubit in marginal])


 

        

    #returns marginalization of an  extended marginal


    return mitigated_marginal_probability_distribution.flatten()




def mitigate_marginals_product(marginals_list: List[Tuple], results_dictionary: Dict[str, Dict[str, int]], noise_model: type[CNModelData] ,ensure_proper_probability_distribution = False )->Dict[str, Dict[tuple[int], np.array]]:
    
    #JT this dictionary 
    mitigated_marginals_dictionary ={}
   
    for marginal in marginals_list:
        mitigated_marginals_dictionary[marginal] = mitigate_marginal_product(marginal=marginal,results_dictionary=results_dictionary,noise_model=noise_model,ensure_proper_probability_distribution=ensure_proper_probability_distribution)
   
    return {next(iter(results_dictionary)): mitigated_marginals_dictionary}
